"""Darker - apply black reformatting to only areas edited since the last commit"""

import concurrent.futures
import logging
import sys
import warnings
from argparse import Action, ArgumentError
from datetime import datetime
from difflib import unified_diff
from pathlib import Path
from typing import Collection, Generator, List, Tuple

from darker.black_diff import (
    BlackConfig,
    filter_python_files,
    read_black_config,
    run_black,
)
from darker.chooser import choose_lines
from darker.command_line import parse_command_line
from darker.concurrency import get_executor
from darker.config import Exclusions, OutputMode, dump_config
from darker.diff import diff_chunks
from darker.exceptions import DependencyError, MissingPackageError
from darker.git import (
    PRE_COMMIT_FROM_TO_REFS,
    WORKTREE,
    EditedLinenumsDiffer,
    RevisionRange,
    get_missing_at_revision,
    get_path_in_repo,
    git_get_content_at_revision,
    git_get_modified_python_files,
    git_is_repository,
)
from darker.help import ISORT_INSTRUCTION
from darker.highlighting import colorize, should_use_color
from darker.import_sorting import apply_isort, isort
from darker.linting import run_linters
from darker.utils import (
    GIT_DATEFORMAT,
    TextDocument,
    debug_dump,
    get_common_root,
    glob_any,
)
from darker.verification import ASTVerifier, BinarySearch, NotEquivalentError

logger = logging.getLogger(__name__)

ProcessedDocument = Tuple[Path, TextDocument, TextDocument]


def format_edited_parts(
    root: Path,
    changed_files: Collection[Path],  # pylint: disable=unsubscriptable-object
    exclude: Exclusions,
    revrange: RevisionRange,
    black_config: BlackConfig,
    report_unmodified: bool,
    workers: int = 1,
) -> Generator[ProcessedDocument, None, None]:
    """Black (and optional isort) formatting modified chunks in a set of files

    Files inside given directories and excluded by Black's configuration are not
    reformatted using Black, but their imports are still sorted. Also, linters will be
    run for all files in a separate step after this function has completed.

    Files listed explicitly on the command line are always reformatted.

    :param root: The common root of all files to reformat
    :param changed_files: Python files and explicitly requested files which have been
                          modified in the repository between the given Git revisions
    :param exclude: Files to exclude when running Black, and when running ``isort``
    :param revrange: The Git revisions to compare
    :param black_config: Configuration to use for running Black
    :param report_unmodified: ``True`` to yield also files which weren't modified
    :param workers: number of cpu processes to use (0 - autodetect)
    :return: A generator which yields details about changes for each file which should
             be reformatted, and skips unchanged files.

    """
    with get_executor(max_workers=workers) as executor:
        # pylint: disable=unsubscriptable-object
        futures: List[concurrent.futures.Future[ProcessedDocument]] = []
        edited_linenums_differ = EditedLinenumsDiffer(root, revrange)
        for relative_path_in_rev2 in sorted(changed_files):
            future = executor.submit(
                _isort_and_blacken_single_file,
                root,
                relative_path_in_rev2,
                edited_linenums_differ,
                exclude,
                revrange,
                black_config,
            )
            futures.append(future)

        for future in concurrent.futures.as_completed(futures):
            (
                absolute_path_in_rev2,
                rev2_content,
                content_after_reformatting,
            ) = future.result()
            if report_unmodified or content_after_reformatting != rev2_content:
                yield (absolute_path_in_rev2, rev2_content, content_after_reformatting)


def _isort_and_blacken_single_file(  # pylint: disable=too-many-arguments
    root: Path,
    relative_path_in_rev2: Path,
    edited_linenums_differ: EditedLinenumsDiffer,
    exclude: Exclusions,
    revrange: RevisionRange,
    black_config: BlackConfig,
) -> ProcessedDocument:
    """Black and/or isort formatting for modified chunks in a single file

    :param root: Root directory for the relative path
    :param relative_path_in_rev2: Relative path to a Python source code file
    :param exclude: Files to exclude when running Black, and when running ``isort``
    :param revrange: The Git revisions to compare
    :param black_config: Configuration to use for running Black
    :return: Details about changes for the file

    """
    # With VSCode, `relative_path_in_rev2` may be a `.py.<HASH>.tmp` file in the
    # working tree instead of a `.py` file.
    absolute_path_in_rev2 = root / relative_path_in_rev2
    rev2_content = git_get_content_at_revision(
        relative_path_in_rev2, revrange.rev2, root
    )
    # 1. run isort
    rev2_isorted = apply_isort(
        rev2_content,
        relative_path_in_rev2,
        exclude.isort,
        edited_linenums_differ,
        black_config.get("config"),
        black_config.get("line_length"),
    )
    has_isort_changes = rev2_isorted != rev2_content
    if not glob_any(relative_path_in_rev2, exclude.black):
        # 9. A re-formatted Python file which produces an identical AST was
        #    created successfully - write an updated file or print the diff if
        #    there were any changes to the original
        content_after_reformatting = _blacken_single_file(
            root,
            relative_path_in_rev2,
            get_path_in_repo(relative_path_in_rev2),
            edited_linenums_differ,
            rev2_content,
            rev2_isorted,
            has_isort_changes,
            black_config,
        )
    else:
        # File was excluded by Black configuration, don't reformat
        content_after_reformatting = rev2_isorted
    return absolute_path_in_rev2, rev2_content, content_after_reformatting


def _blacken_single_file(  # pylint: disable=too-many-arguments,too-many-locals
    root: Path,
    relative_path_in_rev2: Path,
    relative_path_in_repo: Path,
    edited_linenums_differ: EditedLinenumsDiffer,
    rev2_content: TextDocument,
    rev2_isorted: TextDocument,
    has_isort_changes: bool,
    black_config: BlackConfig,
) -> TextDocument:
    """In a Python file, reformat chunks with edits since the last commit using Black

    :param root: The common root of all files to reformat
    :param relative_path_in_rev2: Relative path to a Python source code file. Possibly a
                                  VSCode ``.py.<HASH>.tmp`` file in the working tree.
    :param relative_path_in_repo: Relative path to source in the Git repository. Same as
                                  ``relative_path_in_rev2`` save for VSCode temp files.
    :param edited_linenums_differ: Helper for finding out which lines were edited
    :param rev2_content: Contents of the file at ``revrange.rev2``
    :param rev2_isorted: Contents of the file after optional import sorting
    :param has_isort_changes: ``True`` if ``isort`` was run and modified the file
    :param black_config: Configuration to use for running Black
    :return: Contents of the file after reformatting
    :raise: NotEquivalentError

    """
    absolute_path_in_rev2 = root / relative_path_in_rev2

    # 4. run black
    formatted = run_black(rev2_isorted, black_config)
    logger.debug(
        "Read %s lines from edited file %s",
        len(rev2_isorted.lines),
        absolute_path_in_rev2,
    )
    logger.debug("Black reformat resulted in %s lines", len(formatted.lines))

    # 5. get the diff between the edited and reformatted file
    # 6. convert the diff into chunks
    black_chunks = diff_chunks(rev2_isorted, formatted)

    # Exit early if nothing to do
    if not black_chunks:
        return rev2_isorted

    max_context_lines = len(rev2_isorted.lines)
    minimum_context_lines = BinarySearch(0, max_context_lines + 1)
    last_successful_reformat = None

    verifier = ASTVerifier(baseline=rev2_isorted)

    while not minimum_context_lines.found:
        context_lines = minimum_context_lines.get_next()
        if context_lines > 0:
            logger.debug(
                "Trying with %s lines of context for `git diff -U %s`",
                context_lines,
                absolute_path_in_rev2,
            )
        # 2. diff the given revision and worktree for the file
        # 3. extract line numbers in the edited to-file for changed lines
        edited_linenums = edited_linenums_differ.revision_vs_lines(
            relative_path_in_repo, rev2_isorted, context_lines
        )
        if has_isort_changes and not edited_linenums:
            logger.debug("No changes in %s after isort", absolute_path_in_rev2)
            last_successful_reformat = rev2_isorted
            break

        # 7. choose reformatted content
        chosen = TextDocument.from_lines(
            choose_lines(black_chunks, edited_linenums),
            encoding=rev2_content.encoding,
            newline=rev2_content.newline,
            mtime=datetime.utcnow().strftime(GIT_DATEFORMAT),
        )

        # 8. verify
        logger.debug(
            "Verifying that the %s original edited lines and %s reformatted lines "
            "parse into an identical abstract syntax tree",
            len(rev2_isorted.lines),
            len(chosen.lines),
        )
        if not verifier.is_equivalent_to_baseline(chosen):
            debug_dump(black_chunks, edited_linenums)
            logger.debug(
                "AST verification of %s with %s lines of context failed",
                absolute_path_in_rev2,
                context_lines,
            )
            minimum_context_lines.respond(False)
        else:
            minimum_context_lines.respond(True)
            last_successful_reformat = chosen
    if not last_successful_reformat:
        raise NotEquivalentError(relative_path_in_rev2)
    return last_successful_reformat


def modify_file(path: Path, new_content: TextDocument) -> None:
    """Write new content to a file and inform the user by logging"""
    logger.info("Writing %s bytes into %s", len(new_content.string), path)
    path.write_bytes(new_content.encoded_string)


def print_diff(
    path: Path,
    old: TextDocument,
    new: TextDocument,
    root: Path,
    use_color: bool,
) -> None:
    """Print ``black --diff`` style output for the changes

    :param path: The path of the file to print the diff output for, relative to CWD
    :param old: Old contents of the file
    :param new: New contents of the file
    :param root: The root for the relative path (current working directory if omitted)

    Modification times should be in the format "YYYY-MM-DD HH:MM:SS:mmmmmm +0000"

    """
    relative_path = path.resolve().relative_to(root).as_posix()
    diff = "\n".join(
        line.rstrip("\n")
        for line in unified_diff(
            old.lines,
            new.lines,
            fromfile=relative_path,
            tofile=relative_path,
            fromfiledate=old.mtime,
            tofiledate=new.mtime,
            n=5,  # Black shows 5 lines of context, do the same
        )
    )
    print(colorize(diff, "diff", use_color))


def print_source(new: TextDocument, use_color: bool) -> None:
    """Print the reformatted Python source code"""
    if use_color:
        try:
            (
                highlight,
                TerminalFormatter,  # pylint: disable=invalid-name
                PythonLexer,
            ) = _import_pygments()  # type: ignore
        except ImportError:
            print(new.string, end="")
        else:
            print(highlight(new.string, PythonLexer(), TerminalFormatter()), end="")
    else:
        print(new.string, end="")


def _import_pygments():  # type: ignore
    """Import within a function to ease mocking the import in unit-tests.

    Cannot be typed as it imports parts of its own return type.
    """
    # pylint: disable=import-outside-toplevel
    from pygments import highlight
    from pygments.formatters import (  # pylint: disable=no-name-in-module
        TerminalFormatter,
    )
    from pygments.lexers.python import PythonLexer

    return highlight, TerminalFormatter, PythonLexer


def main(argv: List[str] = None) -> int:
    """Parse the command line and reformat and optionally lint each source file

    1. run isort on each edited file (optional)
    2. diff the given revision and worktree (optionally with isort modifications) for
       all file & dir paths on the command line
    3. extract line numbers in each edited to-file for changed lines
    4. run black on the contents of each edited to-file
    5. get a diff between the edited to-file and the reformatted content
    6. convert the diff into chunks, keeping original and reformatted content for each
       chunk
    7. choose reformatted content for each chunk if there were any changed lines inside
       the chunk in the edited to-file, or choose the chunk's original contents if no
       edits were done in that chunk
    8. verify that the resulting reformatted source code parses to an identical AST as
       the original edited to-file
    9. write the reformatted source back to the original file
    10. run linter subprocesses for all edited files (10.-13. optional)
    11. diff the given revision and worktree (after isort and Black reformatting) for
        each file reported by a linter
    12. extract line numbers in each file reported by a linter for changed lines
    13. print only linter error lines which fall on changed lines

    :param argv: The command line arguments to the ``darker`` command
    :return: 1 if the ``--check`` argument was provided and at least one file was (or
             should be) reformatted; 0 otherwise.

    """
    if argv is None:
        argv = sys.argv[1:]
    args, config, config_nondefault = parse_command_line(argv)
    logging.basicConfig(level=args.log_level)
    if args.log_level == logging.INFO:
        formatter = logging.Formatter("%(levelname)s: %(message)s")
        logging.getLogger().handlers[0].setFormatter(formatter)

    # Make sure we don't get excessive debug log output from Black
    logging.getLogger("blib2to3.pgen2.driver").setLevel(logging.WARNING)

    if args.log_level <= logging.DEBUG:
        print("\n# Effective configuration:\n")
        print(dump_config(config))
        print("\n# Configuration options which differ from defaults:\n")
        print(dump_config(config_nondefault))
        print("\n")

    if args.isort and not isort:
        raise MissingPackageError(f"{ISORT_INSTRUCTION} to use the `--isort` option.")

    black_config = read_black_config(tuple(args.src), args.config)
    if args.config:
        black_config["config"] = args.config
    if args.line_length:
        black_config["line_length"] = args.line_length
    if args.skip_string_normalization is not None:
        black_config["skip_string_normalization"] = args.skip_string_normalization
    if args.skip_magic_trailing_comma is not None:
        black_config["skip_magic_trailing_comma"] = args.skip_magic_trailing_comma

    paths = {Path(p) for p in args.src}
    root = get_common_root(paths)

    revrange = RevisionRange.parse_with_common_ancestor(args.revision, root)
    output_mode = OutputMode.from_args(args)
    write_modified_files = not args.check and output_mode == OutputMode.NOTHING
    if write_modified_files:
        if args.revision == PRE_COMMIT_FROM_TO_REFS and revrange.rev2 == "HEAD":
            warnings.warn(
                "Darker was called by pre-commit, comparing HEAD to an older commit."
                " As an experimental feature, allowing overwriting of files."
                " See https://github.com/akaihola/darker/issues/180 for details."
            )
        elif revrange.rev2 != WORKTREE:
            raise ArgumentError(
                Action(["-r", "--revision"], "revision"),
                f"Can't write reformatted files for revision '{revrange.rev2}'."
                " Either --diff or --check must be used.",
            )

    missing = get_missing_at_revision(paths, revrange.rev2, root)
    if missing:
        missing_reprs = " ".join(repr(str(path)) for path in missing)
        rev2_repr = "the working tree" if revrange.rev2 == WORKTREE else revrange.rev2
        raise ArgumentError(
            Action(["PATH"], "path"),
            f"Error: Path(s) {missing_reprs} do not exist in {rev2_repr}",
        )

    # These are absolute paths:
    files_to_process = filter_python_files(paths, root, {})
    files_to_blacken = filter_python_files(paths, root, black_config)
    if output_mode == OutputMode.CONTENT:
        # With `-d` / `--stdout`, process the file whether modified or not. Paths have
        # previously been validated to contain exactly one existing file.
        changed_files_to_process = {
            p.resolve().relative_to(root) for p in files_to_process
        }
        black_exclude = set()
    else:
        # In other modes, only process files which have been modified.
        if git_is_repository(root):
            changed_files_to_process = git_get_modified_python_files(
                files_to_process, revrange, root
            )
        else:
            changed_files_to_process = {
                path.relative_to(root) for path in files_to_process
            }
        black_exclude = {
            str(path)
            for path in changed_files_to_process
            if root / path not in files_to_blacken
        }

    use_color = should_use_color(config["color"])
    formatting_failures_on_modified_lines = False
    for path, old, new in sorted(
        format_edited_parts(
            root,
            changed_files_to_process,
            Exclusions(black=black_exclude, isort=set() if args.isort else {"**/*"}),
            revrange,
            black_config,
            report_unmodified=output_mode == OutputMode.CONTENT,
            workers=config["workers"],
        ),
    ):
        formatting_failures_on_modified_lines = True
        if output_mode == OutputMode.DIFF:
            print_diff(path, old, new, root, use_color)
        elif output_mode == OutputMode.CONTENT:
            print_source(new, use_color)
        if write_modified_files:
            modify_file(path, new)
    linter_failures_on_modified_lines = run_linters(
        args.lint,
        root,
        changed_files_to_process,
        revrange,
        use_color,
    )
    return (
        1
        if linter_failures_on_modified_lines
        or (args.check and formatting_failures_on_modified_lines)
        else 0
    )


def main_with_error_handling() -> int:
    """Entry point for console script"""
    try:
        return main()
    except (ArgumentError, DependencyError) as exc_info:
        if logger.root.level < logging.WARNING:
            raise
        sys.exit(str(exc_info))


if __name__ == "__main__":
    RETVAL = main_with_error_handling()
    sys.exit(RETVAL)



"""Tests for `darker.multiline_strings`"""

import pytest

from darker import multiline_strings
from darker.utils import TextDocument


def test_get_multiline_string_ranges():
    """`get_multiline_string_ranges()` identifies multi-line strings correctly"""
    content = TextDocument.from_lines(
        line
        for _, line in [
            # Dummy line numbers added to ease debugging of failures
            (1, "'single-quoted string'"),
            (2, "f'single-quoted f-string'"),
            (3, "r'single-quoted raw string'"),
            (4, "rf'single-quoted raw f-string'"),
            (5, "b'single-quoted bytestring'"),
            (6, "rb'single-quoted raw bytestring'"),
            (7, "'''triple-single-quoted one-line string'''"),
            (8, "f'''triple-single-quoted one-line f-string'''"),
            (9, "r'''triple-single-quoted one-line raw string'''"),
            (10, "rf'''triple-single-quoted one-line raw f-string'''"),
            (11, "b'''triple-single-quoted one-line bytestring'''"),
            (12, "rb'''triple-single-quoted one-line raw bytestring'''"),
            (13, "'''triple-single-quoted"),
            (14, "   two-line string'''"),
            (15, "f'''triple-single-quoted"),
            (16, "    two-line f-string'''"),
            (17, "r'''triple-single-quoted"),
            (18, "    two-line raw string'''"),
            (19, "rf'''triple-single-quoted"),
            (20, "     two-line raw f-string'''"),
            (21, "b'''triple-single-quoted"),
            (22, "    two-line bytestring'''"),
            (23, "rb'''triple-single-quoted"),
            (24, "     two-line raw bytestring'''"),
            (25, '"double-quoted string"'),
            (26, 'f"double-quoted f-string"'),
            (27, 'r"double-quoted raw string"'),
            (28, 'rf"double-quoted raw f-string"'),
            (29, 'b"double-quoted bytestring"'),
            (30, 'rb"double-quoted raw bytestring"'),
            (31, '"""triple-double-quoted one-line string"""'),
            (32, 'f"""triple-double-quoted one-line f-string"""'),
            (33, 'r"""triple-double-quoted one-line raw string"""'),
            (34, 'rf"""triple-double-quoted one-line raw f-string"""'),
            (35, 'b"""triple-double-quoted one-line bytestring"""'),
            (36, 'rb"""triple-double-quoted one-line raw bytestring"""'),
            (37, '"""triple-double-quoted'),
            (38, '   two-line string"""'),
            (39, 'f"""triple-double-quoted'),
            (40, '    two-line f-string"""'),
            (41, 'r"""triple-double-quoted'),
            (42, '    two-line raw string"""'),
            (43, 'rf"""triple-double-quoted'),
            (44, '     two-line raw f-string"""'),
            (45, 'b"""triple-double-quoted'),
            (46, '    two-line bytestring"""'),
            (47, 'rb"""triple-double-quoted'),
            (48, '     two-line raw bytestring"""'),
            (49, '"""triple-'),
            (50, "   double-"),
            (51, "   quoted"),
            (52, "   six-"),
            (53, "   line"),
            (54, '   string"""'),
        ]
    )

    result = multiline_strings.get_multiline_string_ranges(content)

    assert result == [
        # 1-based, end-exclusive
        (13, 15),
        (15, 17),
        (17, 19),
        (19, 21),
        (21, 23),
        (23, 25),
        (37, 39),
        (39, 41),
        (41, 43),
        (43, 45),
        (45, 47),
        (47, 49),
        (49, 55),
    ]


# End-exclusive
TEST_RANGES = [(2, 2), (5, 6), (9, 11), (14, 17)]


@pytest.mark.kwparametrize(
    # `(start, end)` and `ranges` are end-exclusive
    dict(start=0, end=0, ranges=[], expect=None),
    dict(start=0, end=42, ranges=[], expect=None),
    dict(start=0, end=0, ranges=TEST_RANGES, expect=None),
    dict(start=1, end=2, ranges=TEST_RANGES, expect=None),
    dict(start=2, end=2, ranges=TEST_RANGES, expect=None),
    dict(start=1, end=3, ranges=TEST_RANGES, expect=(2, 2)),
    dict(start=2, end=3, ranges=TEST_RANGES, expect=None),
    dict(start=3, end=3, ranges=TEST_RANGES, expect=None),
    dict(start=4, end=5, ranges=TEST_RANGES, expect=None),
    dict(start=5, end=5, ranges=TEST_RANGES, expect=None),
    dict(start=4, end=6, ranges=TEST_RANGES, expect=(5, 6)),
    dict(start=5, end=6, ranges=TEST_RANGES, expect=(5, 6)),
    dict(start=6, end=6, ranges=TEST_RANGES, expect=None),
    dict(start=4, end=7, ranges=TEST_RANGES, expect=(5, 6)),
    dict(start=5, end=7, ranges=TEST_RANGES, expect=(5, 6)),
    dict(start=6, end=7, ranges=TEST_RANGES, expect=None),
    dict(start=10, end=10, ranges=TEST_RANGES, expect=(9, 11)),
    dict(start=10, end=11, ranges=TEST_RANGES, expect=(9, 11)),
    dict(start=10, end=12, ranges=TEST_RANGES, expect=(9, 11)),
    dict(start=11, end=11, ranges=TEST_RANGES, expect=None),
    dict(start=11, end=12, ranges=TEST_RANGES, expect=None),
    dict(start=12, end=12, ranges=TEST_RANGES, expect=None),
    dict(start=10, end=19, ranges=TEST_RANGES, expect=(9, 17)),
)
def test_find_overlap(start, end, ranges, expect):
    """`find_overlap()` finds the enclosing range for overlapping ranges"""
    result = multiline_strings.find_overlap(start, end, ranges)

    assert result == expect



"""Command line parsing for the ``darker`` binary"""

from argparse import SUPPRESS, ArgumentParser, Namespace
from typing import Any, List, Optional, Tuple

from darker import help as hlp
from darker.argparse_helpers import (
    LogLevelAction,
    NewlinePreservingFormatter,
    OptionsForReadmeAction,
)
from darker.config import (
    DarkerConfig,
    OutputMode,
    get_effective_config,
    get_modified_config,
    load_config,
    override_color_with_environment,
)
from darker.version import __version__


def make_argument_parser(require_src: bool) -> ArgumentParser:
    """Create the argument parser object

    :param require_src: ``True`` to require at least one path as a positional argument
                        on the command line. ``False`` to not require on.

    """
    parser = ArgumentParser(
        description=hlp.DESCRIPTION, formatter_class=NewlinePreservingFormatter
    )
    parser.register("action", "log_level", LogLevelAction)

    def add_arg(help_text: Optional[str], *name_or_flags: str, **kwargs: Any) -> None:
        kwargs["help"] = help_text
        parser.add_argument(*name_or_flags, **kwargs)

    add_arg(hlp.SRC, "src", nargs="+" if require_src else "*", metavar="PATH")
    add_arg(hlp.REVISION, "-r", "--revision", default="HEAD", metavar="REV")
    add_arg(hlp.DIFF, "--diff", action="store_true")
    add_arg(hlp.STDOUT, "-d", "--stdout", action="store_true")
    add_arg(hlp.CHECK, "--check", action="store_true")
    add_arg(hlp.ISORT, "-i", "--isort", action="store_true")
    add_arg(hlp.LINT, "-L", "--lint", action="append", metavar="CMD", default=[])
    add_arg(hlp.CONFIG, "-c", "--config", metavar="PATH")
    add_arg(
        hlp.VERBOSE,
        "-v",
        "--verbose",
        action="log_level",
        dest="log_level",
        const=-10,
    )
    add_arg(hlp.QUIET, "-q", "--quiet", action="log_level", dest="log_level", const=10)
    add_arg(hlp.COLOR, "--color", action="store_const", dest="color", const=True)
    add_arg(hlp.NO_COLOR, "--no-color", action="store_const", dest="color", const=False)
    add_arg(hlp.VERSION, "--version", action="version", version=__version__)
    add_arg(
        hlp.SKIP_STRING_NORMALIZATION,
        "-S",
        "--skip-string-normalization",
        action="store_const",
        const=True,
    )
    add_arg(
        hlp.NO_SKIP_STRING_NORMALIZATION,
        "--no-skip-string-normalization",
        action="store_const",
        dest="skip_string_normalization",
        const=False,
    )
    add_arg(
        hlp.SKIP_MAGIC_TRAILING_COMMA,
        "--skip-magic-trailing-comma",
        action="store_const",
        dest="skip_magic_trailing_comma",
        const=True,
    )
    add_arg(
        hlp.LINE_LENGTH,
        "-l",
        "--line-length",
        type=int,
        dest="line_length",
        metavar="LENGTH",
    )
    add_arg(hlp.WORKERS, "-W", "--workers", type=int, dest="workers", default=1)
    # A hidden option for printing command lines option in a format suitable for
    # `README.rst`:
    add_arg(SUPPRESS, "--options-for-readme", action=OptionsForReadmeAction)
    return parser


def parse_command_line(argv: List[str]) -> Tuple[Namespace, DarkerConfig, DarkerConfig]:
    """Return the parsed command line, using defaults from a configuration file

    Also return the effective configuration which combines defaults, the configuration
    read from ``pyproject.toml`` (or the path given in ``--config``), environment
    variables, and command line arguments.

    Finally, also return the set of configuration options which differ from defaults.

    """
    # 1. Parse the paths of files/directories to process into `args.src`.
    parser_for_srcs = make_argument_parser(require_src=False)
    args = parser_for_srcs.parse_args(argv)

    # 2. Locate `pyproject.toml` based on the `-c`/`--config` command line option, or
    #    if it's not provided, based on the paths to process, or in the current
    #    directory if no paths were given. Load Darker configuration from it.
    pyproject_config = load_config(args.config, args.src)
    config = override_color_with_environment(pyproject_config)

    # 3. Use configuration as defaults for re-parsing command line arguments, and don't
    #    require file/directory paths if they are specified in configuration.
    parser = make_argument_parser(require_src=not config.get("src"))
    parser.set_defaults(**config)
    args = parser.parse_args(argv)

    # 4. Make sure there aren't invalid option combinations after merging configuration
    #    and command line options.
    OutputMode.validate_diff_stdout(args.diff, args.stdout)
    OutputMode.validate_stdout_src(args.stdout, args.src)

    # 5. Also create a parser which uses the original default configuration values.
    #    This is used to find out differences between the effective configuration and
    #    default configuration values, and print them out in verbose mode.
    parser_with_original_defaults = make_argument_parser(require_src=True)
    return (
        args,
        get_effective_config(args),
        get_modified_config(parser_with_original_defaults, args),
    )



"""Configuration and fixtures for the Pytest based test suite"""

import os
import re
from pathlib import Path
from subprocess import check_call  # nosec
from typing import Dict, Iterable, List, Union

import pytest
from black import find_project_root as black_find_project_root

from darker.git import _git_check_output_lines


class GitRepoFixture:
    def __init__(self, root: Path, env: Dict[str, str]):
        self.root = root
        self.env = env

    @classmethod
    def create_repository(cls, root: Path) -> "GitRepoFixture":
        """Fixture method for creating a Git repository in the given directory"""
        # For testing, ignore ~/.gitconfig settings like templateDir and defaultBranch.
        # Also, this makes sure GIT_DIR or other GIT_* variables are not set, and that
        # Git's messages are in English.
        env = {"HOME": str(root), "LC_ALL": "C", "PATH": os.environ["PATH"]}
        instance = cls(root, env)
        # pylint: disable=protected-access
        instance._run("init", "--initial-branch=master")
        instance._run("config", "user.email", "ci@example.com")
        instance._run("config", "user.name", "CI system")
        return instance

    def _run(self, *args: str) -> None:
        """Helper method to run a Git command line in the repository root"""
        check_call(["git"] + list(args), cwd=self.root, env=self.env)  # nosec

    def _run_and_get_first_line(self, *args: str) -> str:
        """Helper method to run Git in repo root and return first line of output"""
        return _git_check_output_lines(list(args), Path(self.root))[0]

    def add(
        self, paths_and_contents: Dict[str, Union[str, bytes, None]], commit: str = None
    ) -> Dict[str, Path]:
        """Add/remove/modify files and optionally commit the changes

        :param paths_and_contents: Paths of the files relative to repository root, and
                                   new contents for the files as strings. ``None`` can
                                   be specified as the contents in order to remove a
                                   file.
        :param commit: The message for the commit, or ``None`` to skip making a commit.

        """
        absolute_paths = {
            relative_path: self.root / relative_path
            for relative_path in paths_and_contents
        }
        for relative_path, content in paths_and_contents.items():
            path = absolute_paths[relative_path]
            if content is None:
                self._run("rm", "--", relative_path)
                continue
            if isinstance(content, str):
                content = content.encode("utf-8")
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_bytes(content)
            self._run("add", "--", relative_path)
        if commit:
            self._run("commit", "-m", commit)
        return absolute_paths

    def get_hash(self, revision: str = "HEAD") -> str:
        """Return the commit hash at the given revision in the Git repository"""
        return self._run_and_get_first_line("rev-parse", revision)

    def get_branch(self) -> str:
        """Return the active branch name in the Git repository"""
        return self._run_and_get_first_line("symbolic-ref", "--short", "HEAD")

    def create_tag(self, tag_name: str) -> None:
        """Create a tag at current HEAD"""
        self._run("tag", tag_name)

    def create_branch(self, new_branch: str, start_point: str) -> None:
        """Fixture method to create and check out new branch at given starting point"""
        self._run("checkout", "-b", new_branch, start_point)

    def expand_root(self, lines: Iterable[str]) -> List[str]:
        """Replace "{root/<path>}" in strings with the path in the temporary Git repo

        This is used to generate expected strings corresponding to locations of files in
        the temporary Git repository.

        :param lines: The lines of text to process
        :return: Given lines with paths processed

        """
        return [
            re.sub(r"\{root/(.*?)\}", lambda m: str(self.root / str(m.group(1))), line)
            for line in lines
        ]


@pytest.fixture
def git_repo(tmp_path, monkeypatch):
    """Create a temporary Git repository and change current working directory into it"""
    repository = GitRepoFixture.create_repository(tmp_path)
    monkeypatch.chdir(tmp_path)
    # While `GitRepoFixture.create_repository()` already deletes `GIT_*` environment
    # variables for any Git commands run by the fixture, let's explicitly remove
    # `GIT_DIR` in case a test should call Git directly:
    monkeypatch.delenv("GIT_DIR", raising=False)
    return repository


@pytest.fixture
def find_project_root_cache_clear():
    """Clear LRU caching in :func:`black.find_project_root` before each test

    NOTE: We use `darker.black_compat.find_project_root` to wrap Black's original
    function since its signature has changed along the way. However, clearing the cache
    needs to be done on the original of course.

    """
    black_find_project_root.cache_clear()



# pylint: disable=protected-access,too-many-arguments

"""Unit tests for :mod:`darker.linting`"""

import os
import sys
from pathlib import Path
from textwrap import dedent
from unittest.mock import call, patch

import pytest

from darker import linting
from darker.git import WORKTREE, RevisionRange
from darker.tests.helpers import raises_if_exception

SKIP_ON_WINDOWS = [pytest.mark.skip] if sys.platform.startswith("win") else []
SKIP_ON_UNIX = [] if sys.platform.startswith("win") else [pytest.mark.skip]


@pytest.mark.kwparametrize(
    dict(
        line="module.py:42: Just a line number\n",
        expect=(Path("module.py"), 42, "module.py:42:", "Just a line number"),
    ),
    dict(
        line="module.py:42:5: With column  \n",
        expect=(Path("module.py"), 42, "module.py:42:5:", "With column"),
    ),
    dict(
        line="{git_root_absolute}{sep}mod.py:42: Full path\n",
        expect=(
            Path("mod.py"),
            42,
            "{git_root_absolute}{sep}mod.py:42:",
            "Full path",
        ),
    ),
    dict(
        line="{git_root_absolute}{sep}mod.py:42:5: Full path with column\n",
        expect=(
            Path("mod.py"),
            42,
            "{git_root_absolute}{sep}mod.py:42:5:",
            "Full path with column",
        ),
    ),
    dict(
        line="mod.py:42: 123 digits start the description\n",
        expect=(Path("mod.py"), 42, "mod.py:42:", "123 digits start the description"),
    ),
    dict(
        line="mod.py:42:    indented description\n",
        expect=(Path("mod.py"), 42, "mod.py:42:", "   indented description"),
    ),
    dict(
        line="mod.py:42:5:    indented description\n",
        expect=(Path("mod.py"), 42, "mod.py:42:5:", "   indented description"),
    ),
    dict(
        line="nonpython.txt:5: Non-Python file\n",
        expect=(Path("nonpython.txt"), 5, "nonpython.txt:5:", "Non-Python file"),
    ),
    dict(line="mod.py: No line number\n", expect=(Path(), 0, "", "")),
    dict(line="mod.py:foo:5: Invalid line number\n", expect=(Path(), 0, "", "")),
    dict(line="mod.py:42:bar: Invalid column\n", expect=(Path(), 0, "", "")),
    dict(line="/outside/mod.py:5: Outside the repo\n", expect=(Path(), 0, "", "")),
    dict(line="invalid linter output\n", expect=(Path(), 0, "", "")),
    dict(line=" leading:42: whitespace\n", expect=(Path(), 0, "", "")),
    dict(line=" leading:42:5 whitespace and column\n", expect=(Path(), 0, "", "")),
    dict(line="trailing :42: filepath whitespace\n", expect=(Path(), 0, "", "")),
    dict(line="leading: 42: linenum whitespace\n", expect=(Path(), 0, "", "")),
    dict(line="trailing:42 : linenum whitespace\n", expect=(Path(), 0, "", "")),
    dict(line="plus:+42: before linenum\n", expect=(Path(), 0, "", "")),
    dict(line="minus:-42: before linenum\n", expect=(Path(), 0, "", "")),
    dict(line="plus:42:+5 before column\n", expect=(Path(), 0, "", "")),
    dict(line="minus:42:-5 before column\n", expect=(Path(), 0, "", "")),
)
def test_parse_linter_line(git_repo, monkeypatch, line, expect):
    """Linter output is parsed correctly"""
    monkeypatch.chdir(git_repo.root)
    root_abs = git_repo.root.absolute()
    line_expanded = line.format(git_root_absolute=root_abs, sep=os.sep)

    result = linting._parse_linter_line(line_expanded, git_repo.root)

    expect_expanded = (
        expect[0],
        expect[1],
        expect[2].format(git_root_absolute=root_abs, sep=os.sep),
        expect[3],
    )
    assert result == expect_expanded


@pytest.mark.kwparametrize(
    dict(rev2="master", expect=NotImplementedError),
    dict(rev2=WORKTREE, expect=None),
)
def test_require_rev2_worktree(rev2, expect):
    """``_require_rev2_worktree`` raises an exception if rev2 is not ``WORKTREE``"""
    with raises_if_exception(expect):

        linting._require_rev2_worktree(rev2)


def test_check_linter_output():
    """``_check_linter_output()`` runs linter and returns the stdout stream"""
    with linting._check_linter_output(
        "echo", Path("root/of/repo"), {Path("first.py"), Path("second.py")}
    ) as stdout:
        lines = list(stdout)

    assert lines == [
        f"{Path('root/of/repo/first.py')} {Path('root/of/repo/second.py')}\n"
    ]


@pytest.mark.kwparametrize(
    dict(
        _descr="No files to check, no output",
        paths=[],
        location="test.py:1:",
        expect_output=[],
        expect_log=[],
    ),
    dict(
        _descr="Check one file, report on a modified line in test.py",
        paths=["one.py"],
        location="test.py:1:",
        expect_output=["", "test.py:1: {root/one.py}"],
        expect_log=[],
    ),
    dict(
        _descr="Check one file, report on a column of a modified line in test.py",
        paths=["one.py"],
        location="test.py:1:42:",
        expect_output=["", "test.py:1:42: {root/one.py}"],
        expect_log=[],
    ),
    dict(
        _descr="No output if report is on an unmodified line in test.py",
        paths=["one.py"],
        location="test.py:2:42:",
        expect_output=[],
        expect_log=[],
    ),
    dict(
        _descr="No output if report is on a column of an unmodified line in test.py",
        paths=["one.py"],
        location="test.py:2:42:",
        expect_output=[],
        expect_log=[],
    ),
    dict(
        _descr="Check two files, report on a modified line in test.py",
        paths=["one.py", "two.py"],
        location="test.py:1:",
        expect_output=["", "test.py:1: {root/one.py} {root/two.py}"],
        expect_log=[],
    ),
    dict(
        _descr="Check two files, rpeort on a column of a modified line in test.py",
        paths=["one.py", "two.py"],
        location="test.py:1:42:",
        expect_output=["", "test.py:1:42: {root/one.py} {root/two.py}"],
        expect_log=[],
    ),
    dict(
        _descr="No output if 2-file report is on an unmodified line in test.py",
        paths=["one.py", "two.py"],
        location="test.py:2:",
        expect_output=[],
        expect_log=[],
    ),
    dict(
        _descr="No output if 2-file report is on a column of an unmodified line",
        paths=["one.py", "two.py"],
        location="test.py:2:42:",
        expect_output=[],
        expect_log=[],
    ),
    dict(
        _descr="Warning for a file missing from the working tree",
        paths=["missing.py"],
        location="missing.py:1:",
        expect_output=[],
        expect_log=["WARNING Missing file missing.py from echo missing.py:1:"],
    ),
    dict(
        _descr="Linter message for a non-Python file is ignored with a warning",
        paths=["one.py"],
        location="nonpython.txt:1:",
        expect_output=[],
        expect_log=[
            "WARNING Linter message for a non-Python file: "
            "nonpython.txt:1: {root/one.py}"
        ],
    ),
    dict(
        _descr="Message for file outside common root is ignored with a warning (Unix)",
        paths=["one.py"],
        location="/elsewhere/mod.py:1:",
        expect_output=[],
        expect_log=[
            "WARNING Linter message for a file /elsewhere/mod.py "
            "outside requested directory {root/}"
        ],
        marks=SKIP_ON_WINDOWS,
    ),
    dict(
        _descr="Message for file outside common root is ignored with a warning (Win)",
        paths=["one.py"],
        location="C:\\elsewhere\\mod.py:1:",
        expect_output=[],
        expect_log=[
            "WARNING Linter message for a file C:\\elsewhere\\mod.py "
            "outside requested directory {root/}"
        ],
        marks=SKIP_ON_UNIX,
    ),
)
def test_run_linter(
    git_repo, capsys, caplog, _descr, paths, location, expect_output, expect_log
):
    """Linter gets correct paths on command line and outputs just changed lines

    We use ``echo`` as our "linter". It just adds the paths of each file to lint as an
    "error" on a line of ``test.py``. What this test does is the equivalent of e.g.::

    - creating a ``test.py`` such that the first line is modified after the last commit
    - creating and committing ``one.py`` and ``two.py``
    - running::

          $ darker -L 'echo test.py:1:' one.py two.py
          test.py:1: git-repo-root/one.py git-repo-root/two.py

    """
    src_paths = git_repo.add(
        {"test.py": "1\n2\n", "nonpython.txt": "hello\n"}, commit="Initial commit"
    )
    src_paths["test.py"].write_bytes(b"one\n2\n")
    cmdline = f"echo {location}"
    revrange = RevisionRange("HEAD", ":WORKTREE:")

    linting.run_linter(
        cmdline, git_repo.root, {Path(p) for p in paths}, revrange, use_color=False
    )

    # We can now verify that the linter received the correct paths on its command line
    # by checking standard output from the our `echo` "linter".
    # The test cases also verify that only linter reports on modified lines are output.
    result = capsys.readouterr().out.splitlines()
    assert result == git_repo.expand_root(expect_output)
    logs = [f"{record.levelname} {record.message}" for record in caplog.records]
    assert logs == git_repo.expand_root(expect_log)


def test_run_linter_non_worktree():
    """``run_linter()`` doesn't support linting commits, only the worktree"""
    with pytest.raises(NotImplementedError):

        linting.run_linter(
            "dummy-linter",
            Path("/dummy"),
            {Path("dummy.py")},
            RevisionRange.parse_with_common_ancestor("..HEAD", Path("dummy cwd")),
            use_color=False,
        )


@pytest.mark.parametrize(
    "location, expect",
    [
        ("", 0),
        ("test.py:1:", 1),
        ("test.py:2:", 0),
    ],
)
def test_run_linter_return_value(git_repo, location, expect):
    """``run_linter()`` returns the number of linter errors on modified lines"""
    src_paths = git_repo.add({"test.py": "1\n2\n"}, commit="Initial commit")
    src_paths["test.py"].write_bytes(b"one\n2\n")
    cmdline = f"echo {location}"

    result = linting.run_linter(
        cmdline,
        git_repo.root,
        {Path("test.py")},
        RevisionRange("HEAD", ":WORKTREE:"),
        use_color=False,
    )

    assert result == expect


@pytest.mark.kwparametrize(
    dict(
        linter_cmdlines=[],
        linters_return=[],
        expect_result=0,
    ),
    dict(
        linter_cmdlines=["linter"],
        linters_return=[0],
        expect_result=0,
    ),
    dict(
        linter_cmdlines=["linter"],
        linters_return=[1],
        expect_result=1,
    ),
    dict(
        linter_cmdlines=["linter"],
        linters_return=[42],
        expect_result=42,
    ),
    dict(
        linter_cmdlines=["linter1", "linter2"],
        linters_return=[0, 0],
        expect_result=0,
    ),
    dict(
        linter_cmdlines=["linter1", "linter2"],
        linters_return=[0, 42],
        expect_result=42,
    ),
    dict(
        linter_cmdlines=["linter1", "linter2 command line"],
        linters_return=[42, 42],
        expect_result=84,
    ),
)
def test_run_linters(linter_cmdlines, linters_return, expect_result):
    """Unit test for ``run_linters()``"""
    with patch.object(linting, "run_linter") as run_linter:
        run_linter.side_effect = linters_return

        result = linting.run_linters(
            linter_cmdlines,
            Path("dummy root"),
            {Path("dummy paths")},
            RevisionRange("dummy rev1", "dummy rev2"),
            use_color=False,
        )

        expect_calls = [
            call(
                linter_cmdline,
                Path("dummy root"),
                {Path("dummy paths")},
                RevisionRange("dummy rev1", "dummy rev2"),
                False,
            )
            for linter_cmdline in linter_cmdlines
        ]
        assert run_linter.call_args_list == expect_calls
        assert result == expect_result


def test_run_linter_on_new_file(git_repo, capsys):
    """``run_linter()`` considers file missing from history as empty

    Passes through all linter errors as if the original file was empty.

    """
    git_repo.add({"file1.py": "1\n"}, commit="Initial commit")
    git_repo.create_tag("initial")
    (git_repo.root / "file2.py").write_bytes(b"1\n2\n")

    linting.run_linter(
        "echo file2.py:1:",
        Path(git_repo.root),
        {Path("file2.py")},
        RevisionRange("initial", ":WORKTREE:"),
        use_color=False,
    )

    output = capsys.readouterr().out.splitlines()
    assert output == ["", f"file2.py:1: {git_repo.root / 'file2.py'}"]


def test_run_linter_line_separation(git_repo, capsys):
    """``run_linter`` separates contiguous blocks of linter output with empty lines"""
    paths = git_repo.add({"a.py": "1\n2\n3\n4\n5\n6\n"}, commit="Initial commit")
    paths["a.py"].write_bytes(b"a\nb\nc\nd\ne\nf\n")
    linter_output = git_repo.root / "dummy-linter-output.txt"
    linter_output.write_text(
        dedent(
            """
            a.py:2: first block
            a.py:3: of linter output
            a.py:5: second block
            a.py:6: of linter output
            """
        )
    )

    linting.run_linter(
        f"cat {linter_output}",
        Path(git_repo.root),
        {Path(p) for p in paths},
        RevisionRange("HEAD", ":WORKTREE:"),
        use_color=False,
    )

    result = capsys.readouterr().out
    assert result == dedent(
        """
        a.py:2: first block
        a.py:3: of linter output

        a.py:5: second block
        a.py:6: of linter output
        """
    )



# pylint: disable=too-many-arguments,too-many-locals

"""Unit tests for :mod:`darker.command_line` and :mod:`darker.__main__`"""

import os
import re
from argparse import ArgumentError
from importlib import reload
from pathlib import Path
from textwrap import dedent
from unittest.mock import DEFAULT, Mock, call, patch

import pytest
import toml

import darker.help
from darker import black_diff
from darker.__main__ import main
from darker.command_line import make_argument_parser, parse_command_line
from darker.config import ConfigurationError, Exclusions
from darker.git import RevisionRange
from darker.tests.helpers import filter_dict, isort_present, raises_if_exception
from darker.utils import TextDocument, joinlines

pytestmark = pytest.mark.usefixtures("find_project_root_cache_clear")


@pytest.mark.kwparametrize(
    dict(require_src=False, expect=[]), dict(require_src=True, expect=SystemExit)
)
def test_make_argument_parser(require_src, expect):
    """Parser from ``make_argument_parser()`` fails if src required but not provided"""
    parser = make_argument_parser(require_src)
    with raises_if_exception(expect):

        args = parser.parse_args([])

        assert args.src == expect


def get_darker_help_output(capsys):
    """Test for ``--help`` option output"""
    # Make sure the description is re-rendered since its content depends on whether
    # isort is installed or not:
    reload(darker.help)
    with pytest.raises(SystemExit):
        parse_command_line(["--help"])
    return re.sub(r"\s+", " ", capsys.readouterr().out)


@pytest.mark.kwparametrize(
    dict(config=None, argv=[], expect=SystemExit),
    dict(
        config=None,
        argv=["file.py"],
        expect={"src": ["file.py"]},
    ),
    dict(
        config={"src": ["file.py"]},
        argv=[],
        expect={"src": ["file.py"]},
    ),
    dict(
        config={"src": ["file.py"]},
        argv=["file.py"],
        expect={"src": ["file.py"]},
    ),
    dict(
        config={"src": ["file1.py"]},
        argv=["file2.py"],
        expect={"src": ["file2.py"]},
    ),
)
def test_parse_command_line_config_src(
    tmpdir,
    monkeypatch,
    config,
    argv,
    expect,
):
    """The ``src`` positional argument from config and cmdline is handled correctly"""
    monkeypatch.chdir(tmpdir)
    if config is not None:
        toml.dump({"tool": {"darker": config}}, tmpdir / "pyproject.toml")
    with raises_if_exception(expect):

        args, effective_cfg, modified_cfg = parse_command_line(argv)

        assert filter_dict(args.__dict__, "src") == expect
        assert filter_dict(dict(effective_cfg), "src") == expect
        assert filter_dict(dict(modified_cfg), "src") == expect


@pytest.mark.kwparametrize(
    dict(argv=["."], expect="pylint"),
    dict(argv=["./subdir/"], expect="flake8"),
    dict(argv=["--config", "./pyproject.toml", "."], expect="pylint"),
    dict(argv=["--config", "./subdir/pyproject.toml", "."], expect="flake8"),
    dict(argv=["--config", "./pyproject.toml", "subdir/"], expect="pylint"),
    dict(argv=["--config", "./subdir/pyproject.toml", "subdir/"], expect="flake8"),
)
def test_parse_command_line_config_location_specified(
    tmp_path,
    monkeypatch,
    argv,
    expect,
):
    """Darker configuration is read from file pointed to using ``-c``/``--config``"""
    monkeypatch.chdir(tmp_path)
    subdir = tmp_path / "subdir"
    subdir.mkdir()
    root_config = tmp_path / "pyproject.toml"
    subdir_config = subdir / "pyproject.toml"
    root_config.write_text(toml.dumps({"tool": {"darker": {"lint": "pylint"}}}))
    subdir_config.write_text(toml.dumps({"tool": {"darker": {"lint": "flake8"}}}))

    args, effective_cfg, modified_cfg = parse_command_line(argv)

    assert args.lint == expect
    assert effective_cfg["lint"] == expect
    assert modified_cfg["lint"] == expect


@pytest.mark.kwparametrize(
    dict(
        argv=["."],
        expect_value=("src", ["."]),
        expect_config=("src", ["."]),
        expect_modified=("src", ["."]),
    ),
    dict(
        argv=["."],
        expect_value=("revision", "HEAD"),
        expect_config=("revision", "HEAD"),
        expect_modified=("revision", ...),
    ),
    dict(
        argv=["-rmaster", "."],
        expect_value=("revision", "master"),
        expect_config=("revision", "master"),
        expect_modified=("revision", "master"),
    ),
    dict(
        argv=["--revision", "HEAD", "."],
        expect_value=("revision", "HEAD"),
        expect_config=("revision", "HEAD"),
        expect_modified=("revision", ...),
    ),
    dict(
        argv=["."],
        expect_value=("diff", False),
        expect_config=("diff", False),
        expect_modified=("diff", ...),
    ),
    dict(
        argv=["--diff", "."],
        expect_value=("diff", True),
        expect_config=("diff", True),
        expect_modified=("diff", True),
    ),
    dict(
        argv=["."],
        expect_value=("stdout", False),
        expect_config=("stdout", False),
        expect_modified=("stdout", ...),
    ),
    dict(
        argv=["--stdout", "dummy.py"],
        expect_value=("stdout", True),
        expect_config=("stdout", True),
        expect_modified=("stdout", True),
    ),
    dict(
        argv=["--diff", "--stdout", "dummy.py"],
        expect_value=ConfigurationError,
        expect_config=ConfigurationError,
        expect_modified=ConfigurationError,
    ),
    dict(
        argv=["."],
        expect_value=("check", False),
        expect_config=("check", False),
        expect_modified=("check", ...),
    ),
    dict(
        argv=["--check", "."],
        expect_value=("check", True),
        expect_config=("check", True),
        expect_modified=("check", True),
    ),
    dict(
        argv=["."],
        expect_value=("isort", False),
        expect_config=("isort", False),
        expect_modified=("isort", ...),
    ),
    dict(
        argv=["-i", "."],
        expect_value=("isort", True),
        expect_config=("isort", True),
        expect_modified=("isort", True),
    ),
    dict(
        argv=["--isort", "."],
        expect_value=("isort", True),
        expect_config=("isort", True),
        expect_modified=("isort", True),
    ),
    dict(
        argv=["."],
        expect_value=("lint", []),
        expect_config=("lint", []),
        expect_modified=("lint", ...),
    ),
    dict(
        argv=["-L", "pylint", "."],
        expect_value=("lint", ["pylint"]),
        expect_config=("lint", ["pylint"]),
        expect_modified=("lint", ["pylint"]),
    ),
    dict(
        argv=["--lint", "flake8", "-L", "mypy", "."],
        expect_value=("lint", ["flake8", "mypy"]),
        expect_config=("lint", ["flake8", "mypy"]),
        expect_modified=("lint", ["flake8", "mypy"]),
    ),
    dict(
        argv=["."],
        expect_value=("config", None),
        expect_config=("config", None),
        expect_modified=("config", ...),
    ),
    dict(
        argv=["-c", "my.cfg", "."],
        expect_value=("config", "my.cfg"),
        expect_config=("config", "my.cfg"),
        expect_modified=("config", "my.cfg"),
    ),
    dict(
        argv=["--config=my.cfg", "."],
        expect_value=("config", "my.cfg"),
        expect_config=("config", "my.cfg"),
        expect_modified=("config", "my.cfg"),
    ),
    dict(
        argv=["-c", "subdir_with_config", "."],
        expect_value=("config", "subdir_with_config"),
        expect_config=("config", "subdir_with_config"),
        expect_modified=("config", "subdir_with_config"),
    ),
    dict(
        argv=["--config=subdir_with_config", "."],
        expect_value=("config", "subdir_with_config"),
        expect_config=("config", "subdir_with_config"),
        expect_modified=("config", "subdir_with_config"),
    ),
    dict(
        argv=["."],
        expect_value=("log_level", 30),
        expect_config=("log_level", "WARNING"),
        expect_modified=("log_level", ...),
    ),
    dict(
        argv=["-v", "."],
        expect_value=("log_level", 20),
        expect_config=("log_level", "INFO"),
        expect_modified=("log_level", "INFO"),
    ),
    dict(
        argv=["--verbose", "-v", "."],
        expect_value=("log_level", 10),
        expect_config=("log_level", "DEBUG"),
        expect_modified=("log_level", "DEBUG"),
    ),
    dict(
        argv=["-q", "."],
        expect_value=("log_level", 40),
        expect_config=("log_level", "ERROR"),
        expect_modified=("log_level", "ERROR"),
    ),
    dict(
        argv=["--quiet", "-q", "."],
        expect_value=("log_level", 50),
        expect_config=("log_level", "CRITICAL"),
        expect_modified=("log_level", "CRITICAL"),
    ),
    dict(
        argv=["."],
        environ={},
        expect_value=("color", None),
        expect_config=("color", None),
        expect_modified=("color", ...),
    ),
    dict(
        argv=["."],
        environ={"PY_COLORS": "0"},
        expect_value=("color", False),
        expect_config=("color", False),
        expect_modified=("color", False),
    ),
    dict(
        argv=["."],
        environ={"PY_COLORS": "1"},
        expect_value=("color", True),
        expect_config=("color", True),
        expect_modified=("color", True),
    ),
    dict(
        argv=["--color", "."],
        environ={},
        expect_value=("color", True),
        expect_config=("color", True),
        expect_modified=("color", True),
    ),
    dict(
        argv=["--color", "."],
        environ={"PY_COLORS": "0"},
        expect_value=("color", True),
        expect_config=("color", True),
        expect_modified=("color", True),
    ),
    dict(
        argv=["--color", "."],
        environ={"PY_COLORS": "1"},
        expect_value=("color", True),
        expect_config=("color", True),
        expect_modified=("color", True),
    ),
    dict(
        argv=["--no-color", "."],
        environ={},
        expect_value=("color", False),
        expect_config=("color", False),
        expect_modified=("color", False),
    ),
    dict(
        argv=["--no-color", "."],
        environ={"PY_COLORS": "0"},
        expect_value=("color", False),
        expect_config=("color", False),
        expect_modified=("color", False),
    ),
    dict(
        argv=["--no-color", "."],
        environ={"PY_COLORS": "1"},
        expect_value=("color", False),
        expect_config=("color", False),
        expect_modified=("color", False),
    ),
    dict(
        argv=["."],
        expect_value=("skip_string_normalization", None),
        expect_config=("skip_string_normalization", None),
        expect_modified=("skip_string_normalization", ...),
    ),
    dict(
        argv=["-S", "."],
        expect_value=("skip_string_normalization", True),
        expect_config=("skip_string_normalization", True),
        expect_modified=("skip_string_normalization", True),
    ),
    dict(
        argv=["--skip-string-normalization", "."],
        expect_value=("skip_string_normalization", True),
        expect_config=("skip_string_normalization", True),
        expect_modified=("skip_string_normalization", True),
    ),
    dict(
        argv=["--no-skip-string-normalization", "."],
        expect_value=("skip_string_normalization", False),
        expect_config=("skip_string_normalization", False),
        expect_modified=("skip_string_normalization", False),
    ),
    dict(
        argv=["--skip-magic-trailing-comma", "."],
        expect_value=("skip_magic_trailing_comma", True),
        expect_config=("skip_magic_trailing_comma", True),
        expect_modified=("skip_magic_trailing_comma", True),
    ),
    dict(
        argv=["."],
        expect_value=("line_length", None),
        expect_config=("line_length", None),
        expect_modified=("line_length", ...),
    ),
    dict(
        argv=["-l=88", "."],
        expect_value=("line_length", 88),
        expect_config=("line_length", 88),
        expect_modified=("line_length", 88),
    ),
    dict(
        argv=["--line-length", "99", "."],
        expect_value=("line_length", 99),
        expect_config=("line_length", 99),
        expect_modified=("line_length", 99),
    ),
    dict(
        # this is accepted as a path, but would later fail if a file or directory with
        # that funky name doesn't exist
        argv=["--suspicious path"],
        expect_value=("src", ["--suspicious path"]),
        expect_config=("src", ["--suspicious path"]),
        expect_modified=("src", ["--suspicious path"]),
    ),
    dict(
        argv=["valid/path", "another/valid/path"],
        expect_value=("src", ["valid/path", "another/valid/path"]),
        expect_config=("src", ["valid/path", "another/valid/path"]),
        expect_modified=("src", ["valid/path", "another/valid/path"]),
    ),
    environ={},
)
def test_parse_command_line(
    tmp_path, monkeypatch, argv, environ, expect_value, expect_config, expect_modified
):
    """``parse_command_line()`` parses options correctly"""
    monkeypatch.chdir(tmp_path)
    (tmp_path / "dummy.py").touch()
    (tmp_path / "my.cfg").touch()
    (tmp_path / "subdir_with_config").mkdir()
    (tmp_path / "subdir_with_config" / "pyproject.toml").touch()
    with patch.dict(os.environ, environ, clear=True), raises_if_exception(
        expect_value
    ) as expect_exception:

        args, effective_cfg, modified_cfg = parse_command_line(argv)

    if not expect_exception:
        arg_name, expect_arg_value = expect_value
        assert getattr(args, arg_name) == expect_arg_value

        option, expect_config_value = expect_config
        if expect_config_value is ...:
            assert option not in effective_cfg
        else:
            assert effective_cfg[option] == expect_config_value  # type: ignore

        modified_option, expect_modified_value = expect_modified
        if expect_modified_value is ...:
            assert modified_option not in modified_cfg
        else:
            assert (
                modified_cfg[modified_option] == expect_modified_value  # type: ignore
            )


def test_help_description_without_isort_package(capsys):
    """``darker --help`` description shows how to add ``isort`` if it's not present"""
    with isort_present(False):

        assert (
            "Please run `pip install darker[isort]` to enable sorting of import "
            "definitions" in get_darker_help_output(capsys)
        )


def test_help_isort_option_without_isort_package(capsys):
    """``--isort`` option help text shows how to install `isort`` if it's not present"""
    with isort_present(False):

        assert (
            "Please run `pip install darker[isort]` to enable usage of this option."
            in get_darker_help_output(capsys)
        )


def test_help_with_isort_package(capsys):
    """``darker --help`` omits ``isort`` installation instructions if it is installed"""
    with isort_present(True):

        assert "Please run" not in get_darker_help_output(capsys)


@pytest.mark.kwparametrize(
    dict(options=[], expect=call()),
    dict(
        options=["-c", "black.cfg"],
        expect=call(line_length=81, string_normalization=True),
    ),
    dict(
        options=["--config", "black.cfg"],
        expect=call(line_length=81, string_normalization=True),
    ),
    dict(options=["-S"], expect=call(string_normalization=False)),
    dict(
        options=["--skip-string-normalization"], expect=call(string_normalization=False)
    ),
    dict(options=["-l", "90"], expect=call(line_length=90)),
    dict(options=["--line-length", "90"], expect=call(line_length=90)),
    dict(
        options=["-c", "black.cfg", "-S"],
        expect=call(line_length=81, string_normalization=False),
    ),
    dict(
        options=["-c", "black.cfg", "-l", "90"],
        expect=call(line_length=90, string_normalization=True),
    ),
    dict(
        options=["-l", "90", "-S"],
        expect=call(line_length=90, string_normalization=False),
    ),
    dict(
        options=["-c", "black.cfg", "-l", "90", "-S"],
        expect=call(line_length=90, string_normalization=False),
    ),
)
def test_black_options(monkeypatch, tmpdir, git_repo, options, expect):
    """Black options from the command line are passed correctly to Black"""
    monkeypatch.chdir(tmpdir)
    (tmpdir / "pyproject.toml").write("[tool.black]\n")
    (tmpdir / "black.cfg").write(
        dedent(
            """
            [tool.black]
            line-length = 81
            skip-string-normalization = false
            """
        )
    )
    added_files = git_repo.add(
        {"main.py": 'print("Hello World!")\n'}, commit="Initial commit"
    )
    added_files["main.py"].write_bytes(b'print ("Hello World!")\n')
    with patch.object(black_diff, "Mode", wraps=black_diff.Mode) as file_mode_class:

        main(options + [str(path) for path in added_files.values()])

    _, expect_args, expect_kwargs = expect
    file_mode_class.assert_called_once_with(*expect_args, **expect_kwargs)


@pytest.mark.kwparametrize(
    dict(config=[], options=[], expect=call()),
    dict(
        config=[],
        options=["--skip-string-normalization"],
        expect=call(string_normalization=False),
    ),
    dict(
        config=[],
        options=["--no-skip-string-normalization"],
        expect=call(string_normalization=True),
    ),
    dict(
        config=["skip_string_normalization = false"],
        options=[],
        expect=call(string_normalization=True),
    ),
    dict(
        config=["skip_string_normalization = false"],
        options=["--skip-string-normalization"],
        expect=call(string_normalization=False),
    ),
    dict(
        config=["skip_string_normalization = false"],
        options=["--no-skip-string-normalization"],
        expect=call(string_normalization=True),
    ),
    dict(
        config=["skip_string_normalization = true"],
        options=[],
        expect=call(string_normalization=False),
    ),
    dict(
        config=["skip_string_normalization = true"],
        options=["--skip-string-normalization"],
        expect=call(string_normalization=False),
    ),
    dict(
        config=["skip_string_normalization = true"],
        options=["--no-skip-string-normalization"],
        expect=call(string_normalization=True),
    ),
)
def test_black_options_skip_string_normalization(git_repo, config, options, expect):
    """Black string normalization config and cmdline option are combined correctly"""
    added_files = git_repo.add(
        {"main.py": "foo", "pyproject.toml": joinlines(["[tool.black]"] + config)},
        commit="Initial commit",
    )
    added_files["main.py"].write_bytes(b"bar")
    mode_class_mock = Mock(wraps=black_diff.Mode)
    # Speed up tests by mocking `format_str` to skip running Black
    format_str = Mock(return_value="bar")
    with patch.multiple(black_diff, Mode=mode_class_mock, format_str=format_str):

        main(options + [str(path) for path in added_files.values()])

    assert mode_class_mock.call_args_list == [expect]


@pytest.mark.parametrize(
    "config, options, expect",
    [
        ([], [], call()),
        ([], ["--skip-magic-trailing-comma"], call(magic_trailing_comma=False)),
        (["skip_magic_trailing_comma = false"], [], call(magic_trailing_comma=True)),
        (["skip_magic_trailing_comma = true"], [], call(magic_trailing_comma=False)),
    ],
)
def test_black_options_skip_magic_trailing_comma(git_repo, config, options, expect):
    """Black string normalization config and cmdline option are combined correctly"""
    added_files = git_repo.add(
        {"main.py": "foo", "pyproject.toml": joinlines(["[tool.black]"] + config)},
        commit="Initial commit",
    )
    added_files["main.py"].write_bytes(b"a = [1, 2,]")
    mode_class_mock = Mock(wraps=black_diff.Mode)
    # Speed up tests by mocking `format_str` to skip running Black
    format_str = Mock(return_value="a = [1, 2,]")
    with patch.multiple(black_diff, Mode=mode_class_mock, format_str=format_str):

        main(options + [str(path) for path in added_files.values()])

    assert mode_class_mock.call_args_list == [expect]


@pytest.mark.kwparametrize(
    dict(
        options=["a.py"],
        # Expected arguments to the `format_edited_parts()` call.
        # `Path("git_root")` will be replaced with the temporary Git repository root:
        expect=(
            Path("git_root"),
            {Path("a.py")},
            Exclusions(black=set(), isort={"**/*"}),
            RevisionRange("HEAD", ":WORKTREE:"),
            {},
        ),
    ),
    dict(
        options=["--isort", "a.py"],
        expect=(
            Path("git_root"),
            {Path("a.py")},
            Exclusions(black=set(), isort=set()),
            RevisionRange("HEAD", ":WORKTREE:"),
            {},
        ),
    ),
    dict(
        options=["--config", "my.cfg", "a.py"],
        expect=(
            Path("git_root"),
            {Path("a.py")},
            Exclusions(black=set(), isort={"**/*"}),
            RevisionRange("HEAD", ":WORKTREE:"),
            {"config": "my.cfg"},
        ),
    ),
    dict(
        options=["--line-length", "90", "a.py"],
        expect=(
            Path("git_root"),
            {Path("a.py")},
            Exclusions(black=set(), isort={"**/*"}),
            RevisionRange("HEAD", ":WORKTREE:"),
            {"line_length": 90},
        ),
    ),
    dict(
        options=["--skip-string-normalization", "a.py"],
        expect=(
            Path("git_root"),
            {Path("a.py")},
            Exclusions(black=set(), isort={"**/*"}),
            RevisionRange("HEAD", ":WORKTREE:"),
            {"skip_string_normalization": True},
        ),
    ),
    dict(
        options=["--diff", "a.py"],
        expect=(
            Path("git_root"),
            {Path("a.py")},
            Exclusions(black=set(), isort={"**/*"}),
            RevisionRange("HEAD", ":WORKTREE:"),
            {},
        ),
    ),
)
def test_options(git_repo, options, expect):
    """The main engine is called with correct parameters based on the command line

    Executed in a clean directory so Darker's own ``pyproject.toml`` doesn't interfere.

    """
    paths = git_repo.add(
        {"a.py": "1\n", "b.py": "2\n", "my.cfg": ""}, commit="Initial commit"
    )
    paths["a.py"].write_bytes(b"one\n")
    with patch('darker.__main__.format_edited_parts') as format_edited_parts:

        retval = main(options)

    expect = (Path(git_repo.root), expect[1]) + expect[2:]
    format_edited_parts.assert_called_once_with(
        *expect, report_unmodified=False, workers=1
    )
    assert retval == 0


@pytest.mark.kwparametrize(
    dict(check=False, changes=False, lintfail=False),
    dict(check=False, changes=False, lintfail=True, expect_retval=1),
    dict(check=False, changes=True, lintfail=False),
    dict(check=False, changes=True, lintfail=True, expect_retval=1),
    dict(check=True, changes=False, lintfail=False),
    dict(check=True, changes=True, lintfail=True, expect_retval=1),
    expect_retval=0,
)
def test_main_retval(git_repo, check, changes, lintfail, expect_retval):
    """main() return value is correct based on --check, reformatting and linters"""
    git_repo.add({"a.py": ""}, commit="Initial commit")
    format_edited_parts = Mock()
    format_edited_parts.return_value = (
        [
            (
                Path("/dummy.py"),
                TextDocument.from_lines(["old"]),
                TextDocument.from_lines(["new"]),
            )
        ]
        if changes
        else []
    )
    run_linters = Mock(return_value=lintfail)
    check_arg_maybe = ["--check"] if check else []
    with patch.multiple(
        "darker.__main__",
        format_edited_parts=format_edited_parts,
        modify_file=DEFAULT,
        run_linters=run_linters,
    ):

        retval = main(check_arg_maybe + ["a.py"])

    assert retval == expect_retval


def test_main_missing_in_worktree(git_repo):
    """An ``ArgumentError`` is raised if given file is not found on disk"""
    paths = git_repo.add({"a.py": ""}, commit="Add a.py")
    paths["a.py"].unlink()

    with pytest.raises(
        ArgumentError,
        match=re.escape(
            "argument PATH: Error: Path(s) 'a.py' do not exist in the working tree"
        ),
    ):

        main(["a.py"])


def test_main_missing_in_revision(git_repo):
    """An ``ArgumentError`` is raised if given file didn't exist in rev2"""
    paths = git_repo.add({"a.py": ""}, commit="Add a.py")
    git_repo.add({"a.py": None}, commit="Delete a.py")
    paths["a.py"].touch()

    with pytest.raises(
        ArgumentError,
        match=re.escape("argument PATH: Error: Path(s) 'a.py' do not exist in HEAD"),
    ):

        main(["--diff", "--revision", "..HEAD", "a.py"])



"""Unit tests for `darker.black_diff`"""

import re
from dataclasses import dataclass, field
from pathlib import Path
from unittest.mock import ANY, patch

import pytest
import regex

from darker import black_diff
from darker.black_diff import (
    BlackConfig,
    filter_python_files,
    read_black_config,
    run_black,
)
from darker.utils import TextDocument


@dataclass
class RegexEquality:
    """Compare equality to either `re.Pattern` or `regex.Pattern`"""

    pattern: str
    flags: int = field(default=re.UNICODE)

    def __eq__(self, other):
        return (
            other.pattern == self.pattern
            and other.flags & 0x1FF == re.compile(self.pattern).flags | self.flags
        )


@pytest.mark.kwparametrize(
    dict(
        config_path=None, config_lines=["line-length = 79"], expect={"line_length": 79}
    ),
    dict(
        config_path="custom.toml",
        config_lines=["line-length = 99"],
        expect={"line_length": 99},
    ),
    dict(
        config_lines=["skip-string-normalization = true"],
        expect={"skip_string_normalization": True},
    ),
    dict(
        config_lines=["skip-string-normalization = false"],
        expect={"skip_string_normalization": False},
    ),
    dict(
        config_lines=["skip-magic-trailing-comma = true"],
        expect={"skip_magic_trailing_comma": True},
    ),
    dict(
        config_lines=["skip-magic-trailing-comma = false"],
        expect={"skip_magic_trailing_comma": False},
    ),
    dict(config_lines=["target-version = ['py37']"], expect={}),
    dict(config_lines=[r"include = '\.pyi$'"], expect={}),
    dict(
        config_lines=[r"exclude = '\.pyx$'"],
        expect={"exclude": RegexEquality("\\.pyx$")},
    ),
    dict(
        config_lines=["extend-exclude = '''", r"^/setup\.py", r"|^/dummy\.py", "'''"],
        expect={
            "extend_exclude": RegexEquality(
                "(?x)^/setup\\.py\n|^/dummy\\.py\n", re.VERBOSE
            )
        },
    ),
    dict(
        config_lines=["force-exclude = '''", r"^/setup\.py", r"|\.pyc$", "'''"],
        expect={"force_exclude": RegexEquality("(?x)^/setup\\.py\n|\\.pyc$\n")},
    ),
    config_path=None,
)
def test_black_config(tmpdir, config_path, config_lines, expect):
    tmpdir = Path(tmpdir)
    src = tmpdir / "src.py"
    toml = tmpdir / (config_path or "pyproject.toml")
    toml.write_text("[tool.black]\n{}\n".format("\n".join(config_lines)))

    config = read_black_config((str(src),), config_path and str(toml))

    assert config == expect


@pytest.mark.kwparametrize(
    dict(
        expect={
            "none",
            "exclude",
            "extend",
            "force",
            "exclude+extend",
            "exclude+force",
            "extend+force",
            "exclude+extend+force",
        }
    ),
    dict(exclude="exclude", expect={"none", "extend", "force", "extend+force"}),
    dict(extend_exclude="extend", expect={"none", "exclude", "force", "exclude+force"}),
    dict(force_exclude="force", expect={"none", "exclude", "extend", "exclude+extend"}),
    dict(exclude="exclude", extend_exclude="extend", expect={"none", "force"}),
    dict(exclude="exclude", force_exclude="force", expect={"none", "extend"}),
    dict(extend_exclude="extend", force_exclude="force", expect={"none", "exclude"}),
    dict(
        exclude="exclude",
        extend_exclude="extend",
        force_exclude="force",
        expect={"none"},
    ),
    exclude=None,
    extend_exclude=None,
    force_exclude=None,
)
def test_filter_python_files(
    tmp_path, monkeypatch, exclude, extend_exclude, force_exclude, expect
):
    """``filter_python_files()`` skips excluded files correctly"""
    monkeypatch.chdir(tmp_path)
    names = {
        Path(name)
        for name in {
            "none.py",
            "exclude.py",
            "extend.py",
            "force.py",
            "exclude+extend.py",
            "exclude+force.py",
            "extend+force.py",
            "exclude+extend+force.py",
            "none+explicit.py",
            "exclude+explicit.py",
            "extend+explicit.py",
            "force+explicit.py",
            "exclude+extend+explicit.py",
            "exclude+force+explicit.py",
            "extend+force+explicit.py",
            "exclude+extend+force+explicit.py",
        }
    }
    paths = {tmp_path / name for name in names}
    for path in paths:
        path.touch()
    black_config = BlackConfig(
        {
            "exclude": regex.compile(exclude) if exclude else None,
            "extend_exclude": regex.compile(extend_exclude) if extend_exclude else None,
            "force_exclude": regex.compile(force_exclude) if force_exclude else None,
        }
    )
    explicit = {
        Path("none+explicit.py"),
        Path("exclude+explicit.py"),
        Path("extend+explicit.py"),
        Path("force+explicit.py"),
        Path("exclude+extend+explicit.py"),
        Path("exclude+force+explicit.py"),
        Path("extend+force+explicit.py"),
        Path("exclude+extend+force+explicit.py"),
    }

    result = filter_python_files({Path(".")} | explicit, tmp_path, black_config)

    expect_paths = {tmp_path / f"{path}.py" for path in expect} | {
        tmp_path / p for p in explicit
    }
    assert result == expect_paths


@pytest.mark.parametrize("encoding", ["utf-8", "iso-8859-1"])
@pytest.mark.parametrize("newline", ["\n", "\r\n"])
def test_run_black(encoding, newline):
    """Running Black through its Python internal API gives correct results"""
    src = TextDocument.from_lines(
        [f"# coding: {encoding}", "print ( 'touché' )"],
        encoding=encoding,
        newline=newline,
    )

    result = run_black(src, BlackConfig())

    assert result.lines == (
        f"# coding: {encoding}",
        'print("touché")',
    )
    assert result.encoding == encoding
    assert result.newline == newline


@pytest.mark.parametrize("newline", ["\n", "\r\n"])
def test_run_black_always_uses_unix_newlines(newline):
    """Content is always passed to Black with Unix newlines"""
    src = TextDocument.from_str(f"print ( 'touché' ){newline}")
    with patch.object(black_diff, "format_str") as format_str:
        format_str.return_value = 'print("touché")\n'

        _ = run_black(src, BlackConfig())

    format_str.assert_called_once_with("print ( 'touché' )\n", mode=ANY)


def test_run_black_ignores_excludes():
    """Black's exclude configuration is ignored by ``run_black()``"""
    src = TextDocument.from_str("a=1\n")

    result = run_black(
        src,
        BlackConfig(
            {
                "exclude": regex.compile(r".*"),
                "extend_exclude": regex.compile(r".*"),
                "force_exclude": regex.compile(r".*"),
            }
        ),
    )

    assert result.string == "a = 1\n"


@pytest.mark.parametrize(
    "src_content, expect",
    [
        ("", ""),
        ("\n", "\n"),
        ("\r\n", "\r\n"),
        (" ", ""),
        ("\t", ""),
        (" \t", ""),
        (" \t\n", "\n"),
        (" \t\r\n", "\r\n"),
    ],
)
def test_run_black_all_whitespace_input(src_content, expect):
    """All-whitespace files are reformatted correctly"""
    src = TextDocument.from_str(src_content)

    result = run_black(src, BlackConfig())

    assert result.string == expect



"""Helpers for listing modified files and getting unmodified content from Git"""

import logging
import os
import re
import sys
from dataclasses import dataclass
from datetime import datetime
from functools import lru_cache
from pathlib import Path
from subprocess import DEVNULL, PIPE, CalledProcessError, check_output, run  # nosec
from typing import Dict, Iterable, List, Optional, Set, Tuple, Union, overload

from darker.diff import diff_and_get_opcodes, opcodes_to_edit_linenums
from darker.multiline_strings import get_multiline_string_ranges
from darker.utils import GIT_DATEFORMAT, TextDocument

logger = logging.getLogger(__name__)


# Split a revision range into the "from" and "to" revisions and the dots in between.
# Handles these cases:
# <rev>..   <rev>..<rev>   ..<rev>
# <rev>...  <rev>...<rev>  ...<rev>
COMMIT_RANGE_RE = re.compile(r"(.*?)(\.{2,3})(.*)$")


# A colon is an invalid character in tag/branch names. Use that in the special value for
# - denoting the working tree as one of the "revisions" in revision ranges
# - referring to the `PRE_COMMIT_FROM_REF` and `PRE_COMMIT_TO_REF` environment variables
#   for determining the revision range
WORKTREE = ":WORKTREE:"
PRE_COMMIT_FROM_TO_REFS = ":PRE-COMMIT:"


def git_is_repository(path: Path) -> bool:
    """Return ``True`` if ``path`` is inside a Git working tree"""
    try:
        lines = _git_check_output_lines(
            ["rev-parse", "--is-inside-work-tree"], path, exit_on_error=False
        )
        return lines[:1] == ["true"]
    except CalledProcessError as exc_info:
        if exc_info.returncode != 128 or not exc_info.stderr.startswith(
            "fatal: not a git repository"
        ):
            raise
        return False


def git_get_mtime_at_commit(path: Path, revision: str, cwd: Path) -> str:
    """Return the committer date of the given file at the given revision

    :param path: The relative path of the file in the Git repository
    :param revision: The Git revision for which to get the file modification time
    :param cwd: The root of the Git repository

    """
    cmd = ["log", "-1", "--format=%ct", revision, "--", path.as_posix()]
    lines = _git_check_output_lines(cmd, cwd)
    return datetime.utcfromtimestamp(int(lines[0])).strftime(GIT_DATEFORMAT)


def git_get_content_at_revision(path: Path, revision: str, cwd: Path) -> TextDocument:
    """Get unmodified text lines of a file at a Git revision

    :param path: The relative path of the file in the Git repository
    :param revision: The Git revision for which to get the file content, or ``WORKTREE``
                     to get what's on disk right now.
    :param cwd: The root of the Git repository

    """
    if path.is_absolute():
        raise ValueError(
            f"the 'path' parameter must receive a relative path, got {path!r} instead"
        )

    if revision == WORKTREE:
        abspath = cwd / path
        return TextDocument.from_file(abspath)
    cmd = ["show", f"{revision}:./{path.as_posix()}"]
    try:
        return TextDocument.from_bytes(
            _git_check_output(cmd, cwd, exit_on_error=False),
            mtime=git_get_mtime_at_commit(path, revision, cwd),
        )
    except CalledProcessError as exc_info:
        if exc_info.returncode != 128:
            for error_line in exc_info.stderr.splitlines():
                logger.error(error_line)
            raise
        # The file didn't exist at the given revision. Act as if it was an empty
        # file, so all current lines appear as edited.
        return TextDocument()


@dataclass(frozen=True)
class RevisionRange:
    """Represent a range of commits in a Git repository for comparing differences

    ``rev1`` is the "old" revision, and ``rev2``, the "new" revision which should be
    compared against ``rev1``.

    When parsing a revision range expression with triple dots (e.g. ``master...HEAD``),
    the branch point, or common ancestor of the revisions, is used instead of the
    provided ``rev1``. This is useful e.g. when CI is doing a check
    on a feature branch, and there have been commits in the main branch after the branch
    point. Without the ability to compare to the branch point, Darker would suggest
    corrections to formatting on lines changes in the main branch even if those lines
    haven't been touched in the feature branch.

    """

    rev1: str
    rev2: str

    @classmethod
    def parse_with_common_ancestor(
        cls, revision_range: str, cwd: Path
    ) -> "RevisionRange":
        """Convert a range expression to a ``RevisionRange`` object

        If the expression contains triple dots (e.g. ``master...HEAD``), finds the
        common ancestor of the two revisions and uses that as the first revision.

        """
        rev1, rev2, use_common_ancestor = cls._parse(revision_range)
        if use_common_ancestor:
            return cls._with_common_ancestor(rev1, rev2, cwd)
        return cls(rev1, rev2)

    @staticmethod
    def _parse(revision_range: str) -> Tuple[str, str, bool]:
        """Convert a range expression to revisions, using common ancestor if appropriate

        >>> RevisionRange._parse("a..b")
        ('a', 'b', False)
        >>> RevisionRange._parse("a...b")
        ('a', 'b', True)
        >>> RevisionRange._parse("a..")
        ('a', ':WORKTREE:', False)
        >>> RevisionRange._parse("a...")
        ('a', ':WORKTREE:', True)

        """
        if revision_range == PRE_COMMIT_FROM_TO_REFS:
            try:
                return (
                    os.environ["PRE_COMMIT_FROM_REF"],
                    os.environ["PRE_COMMIT_TO_REF"],
                    True,
                )
            except KeyError:
                # Fallback to running against HEAD
                revision_range = "HEAD"
        match = COMMIT_RANGE_RE.match(revision_range)
        if match:
            rev1, range_dots, rev2 = match.groups()
            use_common_ancestor = range_dots == "..."
            return (rev1 or "HEAD", rev2 or WORKTREE, use_common_ancestor)
        return (revision_range or "HEAD", WORKTREE, revision_range not in ["", "HEAD"])

    @classmethod
    def _with_common_ancestor(cls, rev1: str, rev2: str, cwd: Path) -> "RevisionRange":
        """Find common ancestor for revisions and return a ``RevisionRange`` object"""
        rev2_for_merge_base = "HEAD" if rev2 == WORKTREE else rev2
        merge_base_cmd = ["merge-base", rev1, rev2_for_merge_base]
        common_ancestor = _git_check_output_lines(merge_base_cmd, cwd)[0]
        rev1_hash = _git_check_output_lines(["show", "-s", "--pretty=%H", rev1], cwd)[0]
        return cls(rev1 if common_ancestor == rev1_hash else common_ancestor, rev2)


def get_path_in_repo(path: Path) -> Path:
    """Return the relative path to the file in the old revision

    This is usually the same as the relative path on the command line. But in the
    special case of VSCode temporary files (like ``file.py.12345.tmp``), we actually
    want to diff against the corresponding ``.py`` file instead.

    """
    if path.suffixes[-3::2] != [".py", ".tmp"]:
        # The file name is not like `*.py.<HASH>.tmp`. Return it as such.
        return path
    # This is a VSCode temporary file. Drop the hash and the `.tmp` suffix to get the
    # original file name for retrieving the previous revision to diff against.
    path_with_hash = path.with_suffix("")
    return path_with_hash.with_suffix("")


def should_reformat_file(path: Path) -> bool:
    return path.exists() and get_path_in_repo(path).suffix == ".py"


@lru_cache(maxsize=1)
def _make_git_env() -> Dict[str, str]:
    """Create custom minimal environment variables to use when invoking Git

    This makes sure that
    - Git always runs in English
    - ``$PATH`` is preserved (essential on NixOS)
    - the environment is otherwise cleared

    """
    return {"LC_ALL": "C", "PATH": os.environ["PATH"]}


def _git_check_output_lines(
    cmd: List[str], cwd: Path, exit_on_error: bool = True
) -> List[str]:
    """Log command line, run Git, split stdout to lines, exit with 123 on error"""
    return _git_check_output(
        cmd,
        cwd,
        exit_on_error=exit_on_error,
        encoding="utf-8",
    ).splitlines()


@overload
def _git_check_output(
    cmd: List[str], cwd: Path, *, exit_on_error: bool = ..., encoding: None = ...
) -> bytes:
    ...


@overload
def _git_check_output(
    cmd: List[str], cwd: Path, *, exit_on_error: bool = ..., encoding: str
) -> str:
    ...


def _git_check_output(
    cmd: List[str],
    cwd: Path,
    *,
    exit_on_error: bool = True,
    encoding: Optional[str] = None,
) -> Union[str, bytes]:
    """Log command line, run Git, return stdout, exit with 123 on error"""
    logger.debug("[%s]$ git %s", cwd, " ".join(cmd))
    try:
        return check_output(  # nosec
            ["git"] + cmd,
            cwd=str(cwd),
            encoding=encoding,
            stderr=PIPE,
            env=_make_git_env(),
        )
    except CalledProcessError as exc_info:
        if not exit_on_error:
            raise
        if exc_info.returncode != 128:
            sys.stderr.write(exc_info.stderr)
            raise

        # Bad revision or another Git failure. Follow Black's example and return the
        # error status 123.
        for error_line in exc_info.stderr.splitlines():
            logger.error(error_line)
        sys.exit(123)


def _git_exists_in_revision(path: Path, rev2: str, cwd: Path) -> bool:
    """Return ``True`` if the given path exists in the given Git revision

    :param path: The path of the file or directory to check
    :param rev2: The Git revision to look at
    :param cwd: The Git repository root
    :return: ``True`` if the file or directory exists at the revision, or ``False`` if
             it doesn't.

    """
    if (cwd / path).resolve() == cwd.resolve():
        return True
    # Surprise: On Windows, `git cat-file` doesn't work with backslash directory
    # separators in paths. We need to use Posix paths and forward slashes instead.
    cmd = ["git", "cat-file", "-e", f"{rev2}:{path.as_posix()}"]
    logger.debug("[%s]$ %s", cwd, " ".join(cmd))
    result = run(  # nosec
        cmd,
        cwd=str(cwd),
        check=False,
        stderr=DEVNULL,
        env=_make_git_env(),
    )
    return result.returncode == 0


def get_missing_at_revision(paths: Iterable[Path], rev2: str, cwd: Path) -> Set[Path]:
    """Return paths missing in the given revision

    In case of ``WORKTREE``, just check if the files exist on the filesystem instead of
    asking Git.

    :param paths: Paths to check
    :param rev2: The Git revision to look at, or ``WORKTREE`` for the working tree
    :param cwd: The Git repository root
    :return: The set of file or directory paths which are missing in the revision

    """
    if rev2 == WORKTREE:
        return {path for path in paths if not path.exists()}
    return {path for path in paths if not _git_exists_in_revision(path, rev2, cwd)}


def _git_diff_name_only(
    rev1: str, rev2: str, relative_paths: Set[Path], cwd: Path
) -> Set[Path]:
    """Collect names of changed files between commits from Git

    :param rev1: The old commit to compare to
    :param rev2: The new commit to compare, or the string ``":WORKTREE:"`` to compare
                 current working tree to ``rev1``
    :param relative_paths: Relative paths to the files to compare
    :param cwd: The Git repository root
    :return: Relative paths of changed files

    """
    diff_cmd = [
        "diff",
        "--name-only",
        "--relative",
        rev1,
        # rev2 is inserted here if not WORKTREE
        "--",
        *{path.as_posix() for path in relative_paths},
    ]
    if rev2 != WORKTREE:
        diff_cmd.insert(diff_cmd.index("--"), rev2)
    lines = _git_check_output_lines(diff_cmd, cwd)
    return {Path(line) for line in lines}


def _git_ls_files_others(relative_paths: Set[Path], cwd: Path) -> Set[Path]:
    """Collect names of untracked non-excluded files from Git

    This will return those files in ``relative_paths`` which are untracked and not
    excluded by ``.gitignore`` or other Git's exclusion mechanisms.

    :param relative_paths: Relative paths to the files to consider
    :param cwd: The Git repository root
    :return: Relative paths of untracked files

    """
    ls_files_cmd = [
        "ls-files",
        "--others",
        "--exclude-standard",
        "--",
        *{path.as_posix() for path in relative_paths},
    ]
    lines = _git_check_output_lines(ls_files_cmd, cwd)
    return {Path(line) for line in lines}


def git_get_modified_python_files(
    paths: Iterable[Path], revrange: RevisionRange, cwd: Path
) -> Set[Path]:
    """Ask Git for modified and untracked ``*.py`` files

    - ``git diff --name-only --relative <rev> -- <path(s)>``
    - ``git ls-files --others --exclude-standard -- <path(s)>``

    :param paths: Paths to the files to diff
    :param revrange: Git revision range to compare
    :param cwd: The Git repository root
    :return: File names relative to the Git repository root

    """
    relative_paths = {p.resolve().relative_to(cwd) for p in paths}
    changed_paths = _git_diff_name_only(
        revrange.rev1, revrange.rev2, relative_paths, cwd
    )
    if revrange.rev2 == WORKTREE:
        changed_paths.update(_git_ls_files_others(relative_paths, cwd))
    return {path for path in changed_paths if should_reformat_file(cwd / path)}


def _revision_vs_lines(
    root: Path, path_in_repo: Path, rev1: str, content: TextDocument, context_lines: int
) -> List[int]:
    """Return changed line numbers between the given revision and given text content

    The revision range was provided when this `EditedLinenumsDiffer` object was
    created. Content for the given `path_in_repo` at `rev1` of that revision is
    taken from the Git repository, and the provided text content is compared against
    that historical version.

    The actual implementation is here instead of in
    `EditedLinenumsDiffer._revision_vs_lines` so it is accessibe both as a method and as
    a module global for use in the `_compare_revisions` function.

    :param root: Root directory for the relative path `path_in_repo`
    :param path_in_repo: Path of the file to compare, relative to repository root
    :param rev1: The Git revision to compare the on-disk worktree version to
    :param content: The contents to compare to, e.g. from current working tree
    :param context_lines: The number of lines to include before and after a change
    :return: Line numbers of lines changed between the revision and given content

    """
    old = git_get_content_at_revision(path_in_repo, rev1, root)
    edited_opcodes = diff_and_get_opcodes(old, content)
    multiline_string_ranges = get_multiline_string_ranges(content)
    return list(
        opcodes_to_edit_linenums(edited_opcodes, context_lines, multiline_string_ranges)
    )


@lru_cache(maxsize=1)
def _compare_revisions(
    root: Path, path_in_repo: Path, rev1: str, context_lines: int
) -> List[int]:
    """Get line numbers of lines changed between a given revision and the worktree

    Also includes `context_lines` number of extra lines before and after each
    modified line.

    The actual implementation is here instead of in
    `EditedLinenumsDiffer.compare_revisions` because `lru_cache` leaks memory when used
    on methods. See https://stackoverflow.com/q/33672412/15770

    :param root: Root directory for the relative path `path_in_repo`
    :param path_in_repo: Path of the file to compare, relative to repository root
    :param rev1: The Git revision to compare the on-disk worktree version to
    :param context_lines: The number of lines to include before and after a change
    :return: Line numbers of lines changed between the revision and given content

    """
    content = TextDocument.from_file(root / path_in_repo)
    linenums = _revision_vs_lines(root, path_in_repo, rev1, content, context_lines)
    logger.debug(
        "Edited line numbers in %s: %s",
        path_in_repo,
        " ".join(str(n) for n in linenums),
    )
    return linenums


@dataclass(frozen=True)
class EditedLinenumsDiffer:
    """Find out changed lines for a file between given Git revisions"""

    root: Path
    revrange: RevisionRange

    def compare_revisions(self, path_in_repo: Path, context_lines: int) -> List[int]:
        """Get line numbers of lines changed between a given revision and the worktree

        Also includes `context_lines` number of extra lines before and after each
        modified line.

        The actual implementation is in the module global `_compare_revisions` function
        because `lru_cache` leaks memory when used on methods.
        See https://stackoverflow.com/q/33672412/15770

        :param path_in_repo: Path of the file to compare, relative to repository root
        :param context_lines: The number of lines to include before and after a change
        :return: Line numbers of lines changed between the revision and given content

        """
        return _compare_revisions(
            self.root, path_in_repo, self.revrange.rev1, context_lines
        )

    def revision_vs_lines(
        self, path_in_repo: Path, content: TextDocument, context_lines: int
    ) -> List[int]:
        """Return changed line numbers between the given revision and given text content

        The revision range was provided when this `EditedLinenumsDiffer` object was
        created. Content for the given `path_in_repo` at `rev1` of that revision is
        taken from the Git repository, and the provided text content is compared against
        that historical version.

        The actual implementation is in the module global `_revision_vs_lines` function
        so this can be called from both as a method and the module global
        `_compare_revisions` function.

        :param path_in_repo: Path of the file to compare, relative to repository root
        :param content: The contents to compare to, e.g. from current working tree
        :param context_lines: The number of lines to include before and after a change
        :return: Line numbers of lines changed between the revision and given content

        """
        return _revision_vs_lines(
            self.root, path_in_repo, self.revrange.rev1, content, context_lines
        )



"""Helper functions for unit tests"""

import re
import sys
from contextlib import contextmanager, nullcontext
from types import ModuleType
from typing import Any, ContextManager, Dict, List, Optional, Union
from unittest.mock import patch

import pytest
from _pytest.python_api import RaisesContext


def filter_dict(dct: Dict[str, Any], filter_key: str) -> Dict[str, Any]:
    """Return only given keys with their values from a dictionary"""
    return {key: value for key, value in dct.items() if key == filter_key}


def raises_if_exception(expect: Any) -> Union[RaisesContext[Any], ContextManager[None]]:
    """Return a ``pytest.raises`` context manager only if expecting an exception

    If the expected value is not an exception, return a dummy context manager.
    On Python 3.8+ :class:`contextlib.nullcontext` is returned,
    while on older Pythons :class:`contextlib.suppress` is used instead.

    """
    if (isinstance(expect, type) and issubclass(expect, BaseException)) or (
        isinstance(expect, tuple)
        and all(
            isinstance(item, type) and issubclass(item, BaseException)
            for item in expect
        )
    ):
        return pytest.raises(expect)
    if isinstance(expect, BaseException):
        return pytest.raises(type(expect), match=re.escape(str(expect)))
    return nullcontext()


def matching_attrs(obj: BaseException, attrs: List[str]) -> Dict[str, int]:
    """Return object attributes whose name matches one in the given list"""
    return {attname: getattr(obj, attname) for attname in dir(obj) if attname in attrs}


@contextmanager
def raises_or_matches(expect, match_exc_attrs):
    """Helper for matching an expected value or an expected raised exception"""
    if isinstance(expect, BaseException):
        with pytest.raises(type(expect)) as exc_info:
            # The lambda callback should never get called
            yield lambda result: False
        exception_attributes = matching_attrs(exc_info.value, match_exc_attrs)
        expected_attributes = matching_attrs(expect, match_exc_attrs)
        assert exception_attributes == expected_attributes
    else:

        def check(result):
            assert result == expect

        yield check


@contextmanager
def isort_present(present):
    """Context manager to remove or add the `isort` package temporarily for a test"""
    if present:
        # Inject a dummy `isort` package temporarily
        fake_isort_module: Optional[ModuleType] = ModuleType("isort")
        # dummy function required by `import_sorting`:
        fake_isort_module.code = None  # type: ignore
    else:
        # Remove the `isort` package temporarily
        fake_isort_module = None
    with patch.dict(sys.modules, {"isort": fake_isort_module}):
        yield



"""Re-format Python source code using Black

In examples below, a simple two-line snippet is used.
The first line will be reformatted by Black, and the second left intact::

    >>> from pathlib import Path
    >>> from unittest.mock import Mock
    >>> src = Path("dummy/file/path.py")
    >>> src_content = TextDocument.from_lines(
    ...     [
    ...         "for i in range(5): print(i)",
    ...         'print("done")',
    ...     ]
    ... )

First, :func:`run_black` uses Black to reformat the contents of a given file.
Reformatted lines are returned e.g.::

    >>> dst = run_black(src_content, black_config={})
    >>> dst.lines
    ('for i in range(5):', '    print(i)', 'print("done")')

See :mod:`darker.diff` and :mod:`darker.chooser`
for how this result is further processed with:

- :func:`~darker.diff.diff_and_get_opcodes`
  to get a diff of the reformatting
- :func:`~darker.diff.opcodes_to_chunks`
  to split the diff into chunks of original and reformatted content
- :func:`~darker.chooser.choose_lines`
  to reconstruct the source code from original and reformatted chunks
  based on whether reformats touch user-edited lines

"""
import inspect
import logging
import sys
from pathlib import Path
from typing import Collection, Optional, Pattern, Set, Tuple

# `FileMode as Mode` required to satisfy mypy==0.782. Strange.
from black import FileMode as Mode
from black import (
    TargetVersion,
    find_pyproject_toml,
    format_str,
    parse_pyproject_toml,
    re_compile_maybe_verbose,
)
from black.const import DEFAULT_EXCLUDES, DEFAULT_INCLUDES
from black.files import gen_python_files
from black.report import Report

from darker.utils import TextDocument

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

__all__ = ["BlackConfig", "Mode", "run_black"]

logger = logging.getLogger(__name__)


DEFAULT_EXCLUDE_RE = re_compile_maybe_verbose(DEFAULT_EXCLUDES)
DEFAULT_INCLUDE_RE = re_compile_maybe_verbose(DEFAULT_INCLUDES)


class BlackConfig(TypedDict, total=False):
    """Type definition for Black configuration dictionaries"""
    config: str
    exclude: Pattern[str]
    extend_exclude: Pattern[str]
    force_exclude: Pattern[str]
    line_length: int
    skip_string_normalization: bool
    skip_magic_trailing_comma: bool


class BlackModeAttributes(TypedDict, total=False):
    """Type definition for items accepted by ``black.Mode``"""
    target_versions: Set[TargetVersion]
    line_length: int
    string_normalization: bool
    is_pyi: bool
    magic_trailing_comma: bool


def read_black_config(src: Tuple[str, ...], value: Optional[str]) -> BlackConfig:
    """Read the black configuration from ``pyproject.toml``

    :param src: The source code files and directories to be processed by Darker
    :param value: The path of the Black configuration file
    :return: A dictionary of those Black parameters from the configuration file which
             are supported by Darker

    """
    value = value or find_pyproject_toml(src)

    if not value:
        return BlackConfig()

    raw_config = parse_pyproject_toml(value)

    config: BlackConfig = {}
    for key in {
        "line_length",
        "skip_magic_trailing_comma",
        "skip_string_normalization",
    }:
        if key in raw_config:
            config[key] = raw_config[key]  # type: ignore
    for key in {"exclude", "extend_exclude", "force_exclude"}:
        if key in raw_config:
            config[key] = re_compile_maybe_verbose(raw_config[key])  # type: ignore
    return config


def filter_python_files(
    paths: Collection[Path],  # pylint: disable=unsubscriptable-object
    root: Path,
    black_config: BlackConfig,
) -> Set[Path]:
    """Get Python files and explicitly listed files not excluded by Black's config

    :param paths: Relative file/directory paths from CWD to Python sources
    :param root: A common root directory for all ``paths``
    :param black_config: Black configuration which contains the exclude options read
                         from Black's configuration files
    :return: Absolute paths of files which should be reformatted according to
             ``black_config``

    """
    sig = inspect.signature(gen_python_files)
    # those two exist and are required in black>=21.7b1.dev9
    kwargs = dict(verbose=False, quiet=False) if "verbose" in sig.parameters else {}
    absolute_paths = {p.resolve() for p in paths}
    directories = {p for p in absolute_paths if p.is_dir()}
    files = {p for p in absolute_paths if p not in directories}
    files_from_directories = set(
        gen_python_files(
            directories,
            root,
            include=DEFAULT_INCLUDE_RE,
            exclude=black_config.get("exclude", DEFAULT_EXCLUDE_RE),
            extend_exclude=black_config.get("extend_exclude"),
            force_exclude=black_config.get("force_exclude"),
            report=Report(),
            gitignore=None,
            **kwargs,
        )
    )
    return files_from_directories | files


def run_black(src_contents: TextDocument, black_config: BlackConfig) -> TextDocument:
    """Run the black formatter for the Python source code given as a string

    :param src_contents: The source code
    :param black_config: Configuration to use for running Black
    :return: The reformatted content

    """
    # Collect relevant Black configuration options from ``black_config`` in order to
    # pass them to Black's ``format_str()``. File exclusion options aren't needed since
    # at this point we already have a single file's content to work on.
    mode = BlackModeAttributes()
    if "line_length" in black_config:
        mode["line_length"] = black_config["line_length"]
    if "skip_magic_trailing_comma" in black_config:
        mode["magic_trailing_comma"] = not black_config["skip_magic_trailing_comma"]
    if "skip_string_normalization" in black_config:
        # The ``black`` command line argument is
        # ``--skip-string-normalization``, but the parameter for
        # ``black.Mode`` needs to be the opposite boolean of
        # ``skip-string-normalization``, hence the inverse boolean
        mode["string_normalization"] = not black_config["skip_string_normalization"]

    # The custom handling of empty and all-whitespace files below will be unnecessary if
    # https://github.com/psf/black/pull/2484 lands in Black.
    contents_for_black = src_contents.string_with_newline("\n")
    if contents_for_black.strip():
        dst_contents = format_str(contents_for_black, mode=Mode(**mode))
    else:
        dst_contents = "\n" if "\n" in src_contents.string else ""
    return TextDocument.from_str(
        dst_contents,
        encoding=src_contents.encoding,
        override_newline=src_contents.newline,
    )



"""Unit tests for :func:`darker.__main__._blacken_single_file`"""

# pylint: disable=protected-access

from pathlib import Path
from textwrap import dedent

import darker.__main__
from darker.git import EditedLinenumsDiffer, RevisionRange
from darker.utils import TextDocument


def test_blacken_single_file_common_ancestor(git_repo):
    """``_blacken_single_file()`` compares to the common ancestor of ``rev1...rev2``"""
    a_py_initial = dedent(
        """\
        a=1

        b=2

        print( a+b )
        """
    )
    a_py_master = dedent(
        """\
        a=1           # every

        b=2           # line

        print( a+b )  # changed
        """
    )
    a_py_feature = dedent(
        """\
        a= 1  # changed

        b=2

        print( a+b )
        """
    )
    a_py_worktree = dedent(
        """\
        a= 1  # changed

        b=2

        print(a+b )  # changed
        """
    )
    git_repo.add({"a.py": a_py_initial}, commit="Initial commit")
    initial = git_repo.get_hash()
    git_repo.add({"a.py": a_py_master}, commit="on master")
    git_repo.create_branch("feature", initial)
    git_repo.add({"a.py": a_py_feature}, commit="on feature")
    worktree = TextDocument.from_str(a_py_worktree)

    result = darker.__main__._blacken_single_file(
        git_repo.root,
        Path("a.py"),
        Path("a.py"),
        EditedLinenumsDiffer(
            git_repo.root,
            RevisionRange.parse_with_common_ancestor("master...", git_repo.root),
        ),
        rev2_content=worktree,
        rev2_isorted=worktree,
        has_isort_changes=False,
        black_config={},
    )

    assert result.lines == (
        "a = 1  # changed",
        "",
        "b=2",
        "",
        "print(a + b)  # changed",
    )


def test_reformat_single_file_docstring(git_repo):
    """``_reformat_single_file()`` handles docstrings as one contiguous block"""
    initial = dedent(
        '''\
        def docstring_func():
            """
        originally unindented

            originally indented
            """
        '''
    )
    modified = dedent(
        '''\
        def docstring_func():
            """
        originally unindented

            modified and still indented
            """
        '''
    )
    expect = dedent(
        '''\
        def docstring_func():
            """
            originally unindented

                modified and still indented
            """
        '''
    )
    paths = git_repo.add({"a.py": initial}, commit="Initial commit")
    paths["a.py"].write_text(modified)

    result = darker.__main__._blacken_single_file(
        git_repo.root,
        Path("a.py"),
        Path("a.py"),
        EditedLinenumsDiffer(
            git_repo.root,
            RevisionRange.parse_with_common_ancestor("HEAD..", git_repo.root),
        ),
        rev2_content=TextDocument.from_str(modified),
        rev2_isorted=TextDocument.from_str(modified),
        has_isort_changes=False,
        black_config={},
    )

    assert result.lines == tuple(expect.splitlines())



"""Unit tests for :mod:`darker.utils`"""

# pylint: disable=redefined-outer-name, comparison-with-callable

import logging
import os
from pathlib import Path
from textwrap import dedent

import pytest

from darker.utils import (
    TextDocument,
    debug_dump,
    detect_newline,
    get_common_root,
    get_path_ancestry,
    joinlines,
)


@pytest.fixture(params=[TextDocument.from_file, TextDocument.from_bytes])
def textdocument_factory(request):
    """Fixture for a factory function that creates a ``TextDocument``

    The fixture can be parametrized with `(bytes) -> TextDocument` functions
    that take the raw bytes of the document.

    By default, it is parametrized with the ``TextDocument.from_file()`` (for
    which it creates a temporary file) and the ``TextDocument.from_bytes()``
    classmethods.
    """
    if request.param == TextDocument.from_file:

        def factory(content):
            tmp_path = request.getfixturevalue("tmp_path")
            path = tmp_path / "test.py"
            path.write_bytes(content)
            return TextDocument.from_file(path)

        return factory

    return request.param


@pytest.fixture
def textdocument(request, textdocument_factory):
    """Fixture for a ``TextDocument``

    The fixture must be parametrized with the raw bytes of the document.
    """
    return textdocument_factory(request.param)


@pytest.mark.kwparametrize(
    dict(string="", expect="\n"),
    dict(string="\n", expect="\n"),
    dict(string="\r\n", expect="\r\n"),
    dict(string="one line\n", expect="\n"),
    dict(string="one line\r\n", expect="\r\n"),
    dict(string="first line\nsecond line\n", expect="\n"),
    dict(string="first line\r\nsecond line\r\n", expect="\r\n"),
    dict(string="first unix\nthen windows\r\n", expect="\n"),
    dict(string="first windows\r\nthen unix\n", expect="\r\n"),
)
def test_detect_newline(string, expect):
    """``detect_newline()`` gives correct results"""
    result = detect_newline(string)

    assert result == expect


def test_debug_dump(caplog, capsys):
    """darker.utils.debug_dump()"""
    caplog.set_level(logging.DEBUG)
    debug_dump([(1, ("black",), ("chunks",))], [2, 3])
    assert capsys.readouterr().out == (
        dedent(
            """\
            --------------------------------------------------------------------------------
             -   1 black
             +     chunks
            --------------------------------------------------------------------------------
            """
        )
    )


def test_joinlines():
    result = joinlines(("a", "b", "c"))
    assert result == "a\nb\nc\n"


def test_get_common_root_empty():
    """``get_common_root()`` raises a ``ValueError`` if ``paths`` argument is empty"""
    with pytest.raises(ValueError):

        get_common_root([])


def test_get_common_root(tmpdir):
    tmpdir = Path(tmpdir)
    path1 = tmpdir / "a" / "b" / "c" / "d"
    path2 = tmpdir / "a" / "e" / ".." / "b" / "f" / "g"
    path3 = tmpdir / "a" / "h" / ".." / "b" / "i"
    result = get_common_root([path1, path2, path3])
    assert result == tmpdir / "a" / "b"


def test_get_common_root_of_directory(tmpdir):
    tmpdir = Path(tmpdir)
    result = get_common_root([tmpdir])
    assert result == tmpdir


def test_get_path_ancestry_for_directory(tmpdir):
    tmpdir = Path(tmpdir)
    result = list(get_path_ancestry(tmpdir))
    assert result[-1] == tmpdir
    assert result[-2] == tmpdir.parent


def test_get_path_ancestry_for_file(tmpdir):
    tmpdir = Path(tmpdir)
    dummy = tmpdir / "dummy"
    dummy.write_text("dummy")
    result = list(get_path_ancestry(dummy))
    assert result[-1] == tmpdir
    assert result[-2] == tmpdir.parent


@pytest.mark.kwparametrize(
    dict(textdocument=TextDocument(), expect="utf-8"),
    dict(textdocument=TextDocument(encoding="utf-8"), expect="utf-8"),
    dict(textdocument=TextDocument(encoding="utf-16"), expect="utf-16"),
    dict(textdocument=TextDocument.from_str(""), expect="utf-8"),
    dict(textdocument=TextDocument.from_str("", encoding="utf-8"), expect="utf-8"),
    dict(textdocument=TextDocument.from_str("", encoding="utf-16"), expect="utf-16"),
    dict(textdocument=TextDocument.from_lines([]), expect="utf-8"),
    dict(textdocument=TextDocument.from_lines([], encoding="utf-8"), expect="utf-8"),
    dict(textdocument=TextDocument.from_lines([], encoding="utf-16"), expect="utf-16"),
)
def test_textdocument_set_encoding(textdocument, expect):
    """TextDocument.encoding is correct from each constructor"""
    assert textdocument.encoding == expect


@pytest.mark.kwparametrize(
    dict(doc=TextDocument(), expect=""),
    dict(doc=TextDocument(lines=["zéro", "un"]), expect="zéro\nun\n"),
    dict(doc=TextDocument(lines=["zéro", "un"], newline="\n"), expect="zéro\nun\n"),
    dict(
        doc=TextDocument(lines=["zéro", "un"], newline="\r\n"), expect="zéro\r\nun\r\n"
    ),
)
def test_textdocument_string(doc, expect):
    """TextDocument.string respects the newline setting"""
    assert doc.string == expect


@pytest.mark.parametrize("newline", ["\n", "\r\n"])
@pytest.mark.kwparametrize(
    dict(textdocument=TextDocument(), expect=""),
    dict(textdocument=TextDocument(lines=["zéro", "un"])),
    dict(textdocument=TextDocument(string="zéro\nun\n")),
    dict(textdocument=TextDocument(lines=["zéro", "un"], newline="\n")),
    dict(textdocument=TextDocument(string="zéro\nun\n", newline="\n")),
    dict(textdocument=TextDocument(lines=["zéro", "un"], newline="\r\n")),
    dict(textdocument=TextDocument(string="zéro\r\nun\r\n", newline="\r\n")),
    expect="zéro{newline}un{newline}",
)
def test_textdocument_string_with_newline(textdocument, newline, expect):
    """TextDocument.string respects the newline setting"""
    result = textdocument.string_with_newline(newline)

    expected = expect.format(newline=newline)
    assert result == expected


@pytest.mark.kwparametrize(
    dict(encoding="utf-8", newline="\n", expect=b"z\xc3\xa9ro\nun\n"),
    dict(encoding="iso-8859-1", newline="\n", expect=b"z\xe9ro\nun\n"),
    dict(encoding="utf-8", newline="\r\n", expect=b"z\xc3\xa9ro\r\nun\r\n"),
    dict(encoding="iso-8859-1", newline="\r\n", expect=b"z\xe9ro\r\nun\r\n"),
)
def test_textdocument_encoded_string(encoding, newline, expect):
    """TextDocument.encoded_string uses correct encoding and newline"""
    textdocument = TextDocument(
        lines=["zéro", "un"], encoding=encoding, newline=newline
    )

    assert textdocument.encoded_string == expect


@pytest.mark.kwparametrize(
    dict(doc=TextDocument(), expect=()),
    dict(doc=TextDocument(string="zéro\nun\n"), expect=("zéro", "un")),
    dict(doc=TextDocument(string="zéro\nun\n", newline="\n"), expect=("zéro", "un")),
    dict(
        doc=TextDocument(string="zéro\r\nun\r\n", newline="\r\n"), expect=("zéro", "un")
    ),
)
def test_textdocument_lines(doc, expect):
    """TextDocument.lines is correct after parsing a string with different newlines"""
    assert doc.lines == expect


@pytest.mark.kwparametrize(
    dict(
        textdocument=TextDocument.from_str(""),
        expect_lines=(),
        expect_encoding="utf-8",
        expect_newline="\n",
        expect_mtime="",
    ),
    dict(
        textdocument=TextDocument.from_str("", encoding="utf-8"),
        expect_lines=(),
        expect_encoding="utf-8",
        expect_newline="\n",
        expect_mtime="",
    ),
    dict(
        textdocument=TextDocument.from_str("", encoding="iso-8859-1"),
        expect_lines=(),
        expect_encoding="iso-8859-1",
        expect_newline="\n",
        expect_mtime="",
    ),
    dict(
        textdocument=TextDocument.from_str("a\nb\n"),
        expect_lines=("a", "b"),
        expect_encoding="utf-8",
        expect_newline="\n",
        expect_mtime="",
    ),
    dict(
        textdocument=TextDocument.from_str("a\r\nb\r\n"),
        expect_lines=("a", "b"),
        expect_encoding="utf-8",
        expect_newline="\r\n",
        expect_mtime="",
    ),
    dict(
        textdocument=TextDocument.from_str("", mtime="my mtime"),
        expect_lines=(),
        expect_encoding="utf-8",
        expect_newline="\n",
        expect_mtime="my mtime",
    ),
)
def test_textdocument_from_str(
    textdocument, expect_lines, expect_encoding, expect_newline, expect_mtime
):
    """TextDocument.from_str() gets correct content, encoding, newlines and mtime"""
    assert textdocument.lines == expect_lines
    assert textdocument.encoding == expect_encoding
    assert textdocument.newline == expect_newline
    assert textdocument.mtime == expect_mtime


@pytest.mark.kwparametrize(
    dict(textdocument=b'print("touch\xc3\xa9")\n', expect="utf-8"),
    dict(textdocument=b'\xef\xbb\xbfprint("touch\xc3\xa9")\n', expect="utf-8-sig"),
    dict(textdocument=b'# coding: iso-8859-1\n"touch\xe9"\n', expect="iso-8859-1"),
    indirect=["textdocument"],
)
def test_textdocument_detect_encoding(textdocument, expect):
    """TextDocument.from_file/bytes() detects the file encoding correctly"""
    assert textdocument.encoding == expect


@pytest.mark.kwparametrize(
    dict(textdocument=b'print("unix")\n', expect="\n"),
    dict(textdocument=b'print("windows")\r\n', expect="\r\n"),
    indirect=["textdocument"],
)
def test_textdocument_detect_newline(textdocument, expect):
    """TextDocument.from_file/bytes() detects the newline sequence correctly"""
    assert textdocument.newline == expect


@pytest.mark.kwparametrize(
    dict(doc1=TextDocument(lines=["foo"]), doc2=TextDocument(lines=[]), expect=False),
    dict(doc1=TextDocument(lines=[]), doc2=TextDocument(lines=["foo"]), expect=False),
    dict(
        doc1=TextDocument(lines=["foo"]), doc2=TextDocument(lines=["bar"]), expect=False
    ),
    dict(
        doc1=TextDocument(lines=["line1", "line2"]),
        doc2=TextDocument(lines=["line1", "line2"]),
        expect=True,
    ),
    dict(
        doc1=TextDocument(lines=["line1", "line2"], encoding="utf-16", newline="\r\n"),
        doc2=TextDocument(lines=["line1", "line2"]),
        expect=True,
    ),
    dict(doc1=TextDocument(lines=["foo"]), doc2=TextDocument(""), expect=False),
    dict(doc1=TextDocument(lines=[]), doc2=TextDocument("foo\n"), expect=False),
    dict(doc1=TextDocument(lines=["foo"]), doc2=TextDocument("bar\n"), expect=False),
    dict(
        doc1=TextDocument(lines=["line1", "line2"]),
        doc2=TextDocument("line1\nline2\n"),
        expect=True,
    ),
    dict(doc1=TextDocument("foo\n"), doc2=TextDocument(lines=[]), expect=False),
    dict(doc1=TextDocument(""), doc2=TextDocument(lines=["foo"]), expect=False),
    dict(doc1=TextDocument("foo\n"), doc2=TextDocument(lines=["bar"]), expect=False),
    dict(
        doc1=TextDocument("line1\nline2\n"),
        doc2=TextDocument(lines=["line1", "line2"]),
        expect=True,
    ),
    dict(doc1=TextDocument("foo\n"), doc2=TextDocument(""), expect=False),
    dict(doc1=TextDocument(""), doc2=TextDocument("foo\n"), expect=False),
    dict(doc1=TextDocument("foo\n"), doc2=TextDocument("bar\n"), expect=False),
    dict(
        doc1=TextDocument("line1\nline2\n"),
        doc2=TextDocument("line1\nline2\n"),
        expect=True,
    ),
    dict(
        doc1=TextDocument("line1\r\nline2\r\n"),
        doc2=TextDocument("line1\nline2\n"),
        expect=True,
    ),
    dict(doc1=TextDocument("foo"), doc2="line1\nline2\n", expect=NotImplemented),
)
def test_textdocument_eq(doc1, doc2, expect):
    """TextDocument.__eq__()"""
    result = doc1.__eq__(doc2)

    assert result == expect


@pytest.mark.kwparametrize(
    dict(document=TextDocument(""), expect="TextDocument([0 lines])"),
    dict(document=TextDocument(lines=[]), expect="TextDocument([0 lines])"),
    dict(document=TextDocument("One line\n"), expect="TextDocument([1 lines])"),
    dict(document=TextDocument(lines=["One line"]), expect="TextDocument([1 lines])"),
    dict(document=TextDocument("Two\nlines\n"), expect="TextDocument([2 lines])"),
    dict(
        document=TextDocument(lines=["Two", "lines"]), expect="TextDocument([2 lines])"
    ),
    dict(
        document=TextDocument(mtime="some mtime"),
        expect="TextDocument([0 lines], mtime='some mtime')",
    ),
    dict(
        document=TextDocument(encoding="utf-8"),
        expect="TextDocument([0 lines])",
    ),
    dict(
        document=TextDocument(encoding="a non-default encoding"),
        expect="TextDocument([0 lines], encoding='a non-default encoding')",
    ),
    dict(
        document=TextDocument(newline="\n"),
        expect="TextDocument([0 lines])",
    ),
    dict(
        document=TextDocument(newline="a non-default newline"),
        expect="TextDocument([0 lines], newline='a non-default newline')",
    ),
)
def test_textdocument_repr(document, expect):
    """TextDocument.__repr__()"""
    result = document.__repr__()

    assert result == expect


@pytest.mark.kwparametrize(
    dict(document=TextDocument(), expect=""),
    dict(document=TextDocument(mtime=""), expect=""),
    dict(document=TextDocument(mtime="dummy mtime"), expect="dummy mtime"),
)
def test_textdocument_mtime(document, expect):
    """TextDocument.mtime"""
    assert document.mtime == expect


def test_textdocument_from_file(tmp_path):
    """TextDocument.from_file()"""
    dummy_txt = tmp_path / "dummy.txt"
    dummy_txt.write_bytes(b"# coding: iso-8859-1\r\ndummy\r\ncontent\r\n")
    os.utime(dummy_txt, (1_000_000_000, 1_000_000_000))

    document = TextDocument.from_file(dummy_txt)

    assert document.string == "# coding: iso-8859-1\r\ndummy\r\ncontent\r\n"
    assert document.lines == ("# coding: iso-8859-1", "dummy", "content")
    assert document.encoding == "iso-8859-1"
    assert document.newline == "\r\n"
    assert document.mtime == "2001-09-09 01:46:40.000000 +0000"



"""Concurrency helpers for enhancing the performance of Darker"""

import sys
from concurrent.futures import Executor, Future, ProcessPoolExecutor
from typing import Any, Callable, Generic, TypeVar, cast

T = TypeVar("T")  # pylint: disable=invalid-name

if sys.version_info < (3, 9):

    class FutureType(Generic[T]):
        """For Python <3.9 compatibility"""

        def set_exception(self, exc_info: BaseException) -> None:
            "Dummy method for typing"

        def set_result(self, result: Any) -> None:
            "Dummy method for typing"

else:
    FutureType = Future


class DummyExecutor(Executor):
    """Dummy synchronous executor to use with ``--workers=1``

    This makes it easier to write test cases for `darker.__main__.main`.

    """

    # pylint: disable=arguments-differ,unsubscriptable-object,broad-except
    def submit(  # type: ignore[override]
        self, fn: Callable[..., T], *args: Any, **kwargs: Any
    ) -> FutureType[T]:
        """Submits "a callable to be executed with the given arguments.

        Executes the callable immediately as ``fn(*args, **kwargs)`` and returns a
        `Future` instance representing the execution of the callable.

        :param fn: The callable to call
        :param args: Positional arguments for the callable
        :param kwargs: Keyword arguments for the callable
        :return: A `Future` representing the given call

        """
        future = cast(FutureType[T], Future())
        try:
            result = fn(*args, **kwargs)
        except BaseException as exc_info:
            future.set_exception(exc_info)
        else:
            future.set_result(result)
        return future


def get_executor(max_workers: int) -> Executor:
    """Return either a dummy executor (if ``max_workers===1`) or a process pool executor

    :param max_workers: The maximum number of processes that can be used to execute the
                        given calls. If ``0`` then as many worker processes will be
                        created as the machine has processors. If ``1``, the dummy
                        executor will be used so calls are executed synchronously.
    :return: A dummy executor or a process pool executor

    """
    return (
        DummyExecutor()
        if max_workers == 1
        else ProcessPoolExecutor(max_workers or None)
    )



"""Verification for unchanged AST before and after reformatting"""

from typing import Dict, List

from black import assert_equivalent, parse_ast, stringify_ast

from darker.utils import DiffChunk, TextDocument, debug_dump


class NotEquivalentError(Exception):
    pass


class BinarySearch:
    """Effectively search the first index for which a condition is ``True``

    Example usage to find first number between 0 and 100 whose square is larger than
    1000:

    >>> s = BinarySearch(0, 100)
    >>> while not s.found:
    ...     s.respond(s.get_next() ** 2 > 1000)
    >>> assert s.result == 32

    """

    def __init__(self, low: int, high: int):
        self.low = self.mid = low
        self.high = high

    def get_next(self) -> int:
        """Get the next integer index to try"""
        return self.mid

    def respond(self, value: bool) -> None:
        """Provide a ``False`` or ``True`` answer for the current integer index"""
        if value:
            self.high = self.mid
        else:
            self.low = self.mid + 1
        self.mid = (self.low + self.high) // 2

    @property
    def found(self) -> bool:
        """Return ``True`` if the lowest ``True`` response index has been found"""
        return self.low >= self.high

    @property
    def result(self) -> int:
        """Return the lowest index with a ``True`` response"""
        if not self.found:
            raise RuntimeError(
                "Trying to get binary search result before search has been exhausted"
            )
        return self.high


def verify_ast_unchanged(
    edited_to_file: TextDocument,
    reformatted: TextDocument,
    black_chunks: List[DiffChunk],
    edited_linenums: List[int],
) -> None:
    """Verify that source code parses to the same AST before and after reformat"""
    try:
        assert_equivalent(edited_to_file.string, reformatted.string)
    except AssertionError as exc_info:
        debug_dump(black_chunks, edited_linenums)
        raise NotEquivalentError(str(exc_info))


class ASTVerifier:  # pylint: disable=too-few-public-methods
    """Verify if reformatted TextDocument is AST-equivalent to baseline

    Keeps in-memory data about previous comparisons to improve performance.

    """

    def __init__(self, baseline: TextDocument) -> None:
        self._baseline_ast_str = self._to_ast_str(baseline)
        self._comparisons: Dict[str, bool] = {baseline.string: True}

    @staticmethod
    def _to_ast_str(document: TextDocument) -> str:
        return "\n".join(stringify_ast(parse_ast(document.string)))

    def is_equivalent_to_baseline(self, document: TextDocument) -> bool:
        """Returns true if document is AST-equivalent to baseline"""
        if document.string in self._comparisons:
            return self._comparisons[document.string]

        try:
            document_ast_str = self._to_ast_str(document)
        except SyntaxError:
            comparison = False
        else:
            comparison = self._baseline_ast_str == document_ast_str

        self._comparisons[document.string] = comparison
        return self._comparisons[document.string]



"""Helper functions for dealing with reformatting multi-line strings"""

from tokenize import STRING, tokenize
from typing import List, Optional, Sequence, Tuple

from darker.utils import TextDocument


def get_multiline_string_ranges(content: TextDocument) -> List[Tuple[int, int]]:
    """Return the line ranges of multi-line strings found in the given Python source

    The returned data is 1-based, end-exclusive. In other words, each item is a 1-based
    ``(start, end)`` tuple, ``end`` being the line number following the last line of the
    multi-line string.

    :return: Line number ranges of multi-line strings

    """
    readline = (f"{line}\n".encode(content.encoding) for line in content.lines).__next__
    return [
        (t.start[0], t.end[0] + 1)
        for t in tokenize(readline)
        if t.type == STRING and t.end[0] > t.start[0]
    ]


def find_overlap(
    start: int, end: int, ranges: Sequence[Tuple[int, int]]
) -> Optional[Tuple[int, int]]:
    """Return the convex hull of given ranges which overlap with given start and end

    `start..end` must be end-exclusive.

    :param start: The first line of the range to find overlaps for.
    :param end: The first line after the range to find overlaps for.
    :param ranges: The ranges to scan when looking for an overlap with `(start, end)`.
                   End-exclusive, and either 0- or 1-based as long as similarly based as
                   `(start, end)`.
    :return: The convex hull, i.e. range from start of first until the end of last range
             which overlap with the given `(start, end)` range

    """
    overlap: Optional[Tuple[int, int]] = None
    for range_start, range_end in ranges:
        if end <= range_start:
            break
        if start < range_end:
            if overlap:
                overlap = (
                    overlap[0],  # pylint: disable=unsubscriptable-object
                    range_end,
                )
            else:
                overlap = range_start, range_end
    return overlap



"""Tests for the ``darker.argparse_helpers`` module"""

from argparse import ArgumentParser, Namespace
from logging import CRITICAL, DEBUG, ERROR, INFO, NOTSET, WARNING

import pytest

from darker import argparse_helpers
from darker.tests.helpers import raises_if_exception


@pytest.mark.kwparametrize(
    dict(line="", width=0, expect=ValueError),
    dict(line="", width=1, expect=[]),
    dict(
        line="lorem ipsum dolor sit amet",
        width=9,
        expect=["    lorem", "    ipsum", "    dolor", "    sit", "    amet"],
    ),
    dict(
        line="lorem ipsum dolor sit amet",
        width=15,
        expect=["    lorem ipsum", "    dolor sit", "    amet"],
    ),
)
def test_fill_line(line, width, expect):
    """``_fill_line()`` wraps lines correctly"""
    with raises_if_exception(expect):

        result = argparse_helpers._fill_line(  # pylint: disable=protected-access
            line, width, indent="    "
        )

        assert result.splitlines() == expect


@pytest.mark.kwparametrize(
    dict(
        text="lorem ipsum dolor sit amet",
        expect=["    lorem ipsum", "    dolor sit", "    amet"],
    ),
    dict(
        text="lorem\nipsum dolor sit amet",
        expect=["    lorem", "    ipsum dolor", "    sit amet"],
    ),
)
def test_newline_preserving_formatter(text, expect):
    """``NewlinePreservingFormatter`` wraps lines and keeps newlines correctly"""
    formatter = argparse_helpers.NewlinePreservingFormatter("dummy")

    result = formatter._fill_text(  # pylint: disable=protected-access
        text, width=15, indent="    "
    )

    assert result.splitlines() == expect


@pytest.mark.kwparametrize(
    dict(const=10, initial=NOTSET, expect=DEBUG),
    dict(const=10, initial=DEBUG, expect=INFO),
    dict(const=10, initial=INFO, expect=WARNING),
    dict(const=10, initial=WARNING, expect=ERROR),
    dict(const=10, initial=ERROR, expect=CRITICAL),
    dict(const=10, initial=CRITICAL, expect=CRITICAL),
    dict(const=-10, initial=DEBUG, expect=DEBUG),
    dict(const=-10, initial=INFO, expect=DEBUG),
    dict(const=-10, initial=WARNING, expect=INFO),
    dict(const=-10, initial=ERROR, expect=WARNING),
    dict(const=-10, initial=CRITICAL, expect=ERROR),
)
def test_log_level_action(const, initial, expect):
    """``LogLevelAction`` increments/decrements the log level value correctly"""
    action = argparse_helpers.LogLevelAction([], "log_level", const)
    parser = ArgumentParser()
    namespace = Namespace()
    namespace.log_level = initial

    action(parser, namespace, [])

    assert namespace.log_level == expect


@pytest.mark.kwparametrize(
    dict(const=10, count=NOTSET, expect=WARNING),
    dict(const=10, count=1, expect=ERROR),
    dict(const=10, count=2, expect=CRITICAL),
    dict(const=10, count=3, expect=CRITICAL),
    dict(const=-10, count=NOTSET, expect=WARNING),
    dict(const=-10, count=1, expect=INFO),
    dict(const=-10, count=2, expect=DEBUG),
    dict(const=-10, count=3, expect=DEBUG),
)
def test_argumentparser_log_level_action(const, count, expect):
    """The log level action works correctly with an ``ArgumentParser``"""
    parser = ArgumentParser()
    parser.register("action", "log_level", argparse_helpers.LogLevelAction)
    parser.add_argument("-l", dest="log_level", action="log_level", const=const)

    args = parser.parse_args(count * ["-l"])

    assert args.log_level == expect



"""Run arbitrary linters on given files in a Git repository

This supports any linter which reports on standard output and has a fairly standard
correct output format::

    <path>:<linenum>: <description>
    <path>:<linenum>:<column>: <description>

For example, Mypy outputs::

    module.py:57: error: Function is missing a type annotation for one or more arguments

Pylint, on the other hand::

    module.py:44:8: R1720: Unnecessary "elif" after "raise" (no-else-raise)

All such output from the linter will be printed on the standard output
provided that the ``<linenum>`` falls on a changed line.

"""

import logging
from contextlib import contextmanager
from pathlib import Path
from subprocess import PIPE, Popen  # nosec
from typing import IO, Generator, List, Set, Tuple

from darker.git import WORKTREE, EditedLinenumsDiffer, RevisionRange
from darker.highlighting import colorize

logger = logging.getLogger(__name__)


def _strict_nonneg_int(text: str) -> int:
    """Strict parsing of strings to non-negative integers

    Allow no leading or trailing whitespace, nor plus or minus signs.

    :param text: The string to convert
    :raises ValueError: Raises if the string has any non-numeric characters
    :return: [description]
    :rtype: [type]
    """
    if text.strip("+-\t ") != text:
        raise ValueError(r"invalid literal for int() with base 10: {text}")
    return int(text)


def _parse_linter_line(line: str, root: Path) -> Tuple[Path, int, str, str]:
    """Parse one line of linter output

    Only parses lines with
    - a file path (without leading-trailing whitespace),
    - a non-negative line number (without leading/trailing whitespace),
    - optionally a column number (without leading/trailing whitespace), and
    - a description.

    Examples of successfully parsed lines::

        path/to/file.py:42: Description
        path/to/file.py:42:5: Description

    Given a root of ``Path("path/")``, these would be parsed into::

        (Path("to/file.py"), 42, "path/to/file.py:42:", "Description")
        (Path("to/file.py"), 42, "path/to/file.py:42:5:", "Description")

    For all other lines, a dummy entry is returned: an empty path, zero as the line
    number, an empty location string and an empty description. Such lines should be
    simply ignored, since many linters display supplementary information insterspersed
    with the actual linting notifications.

    :param line: The linter output line to parse. May have a trailing newline.
    :param root: The root directory to resolve full file paths against
    :return: A 4-tuple of
             - a ``root``-relative file path,
             - the line number,
             - the path and location string, and
             - the description.

    """
    try:
        location, description = line.rstrip().split(": ", 1)
        if location[1:3] == ":\\":
            # Absolute Windows paths need special handling. Separate out the ``C:`` (or
            # similar), then split by colons, and finally re-insert the ``C:``.
            path_in_drive, linenum_str, *rest = location[2:].split(":")
            path_str = f"{location[:2]}{path_in_drive}"
        else:
            path_str, linenum_str, *rest = location.split(":")
        if path_str.strip() != path_str:
            raise ValueError(r"Filename {path_str!r} has leading/trailing whitespace")
        linenum = _strict_nonneg_int(linenum_str)
        if len(rest) > 1:
            raise ValueError("Too many colon-separated tokens in {location!r}")
        if len(rest) == 1:
            # Make sure it column looks like an int on "<path>:<linenum>:<column>"
            _column = _strict_nonneg_int(rest[0])  # noqa: F841
    except ValueError:
        # Encountered a non-parsable line which doesn't express a linting error.
        # For example, on Mypy:
        # "Found XX errors in YY files (checked ZZ source files)"
        # "Success: no issues found in 1 source file"
        logger.debug("Unparsable linter output: %s", line[:-1])
        return Path(), 0, "", ""
    path_from_cwd = Path(path_str).absolute()
    try:
        path_in_repo = path_from_cwd.relative_to(root)
    except ValueError:
        logger.warning(
            "Linter message for a file %s outside requested directory %s",
            path_from_cwd,
            root,
        )
        return Path(), 0, "", ""
    return path_in_repo, linenum, location + ":", description


def _require_rev2_worktree(rev2: str) -> None:
    """Exit with an error message if ``rev2`` is not ``WORKTREE``

    This is used when running linters since linting arbitrary commits is not supported.

    :param rev2: The ``rev2`` revision to lint.

    """
    if rev2 != WORKTREE:
        raise NotImplementedError(
            "Linting arbitrary commits is not supported. "
            "Please use -r {<rev>|<rev>..|<rev>...} instead."
        )


@contextmanager
def _check_linter_output(
    cmdline: str, root: Path, paths: Set[Path]
) -> Generator[IO[str], None, None]:
    """Run a linter as a subprocess and return its standard output stream

    :param cmdline: The command line for running the linter
    :param root: The common root of all files to lint
    :param paths: Paths of files to check, relative to ``git_root``
    :return: The standard output stream of the linter subprocess

    """
    cmdline_and_paths = cmdline.split() + [str(root / path) for path in sorted(paths)]
    logger.debug("[%s]$ %s", Path.cwd(), " ".join(cmdline_and_paths))
    with Popen(  # nosec
        cmdline_and_paths,
        stdout=PIPE,
        encoding="utf-8",
    ) as linter_process:
        # condition needed for MyPy (see https://stackoverflow.com/q/57350490/15770)
        if linter_process.stdout is None:
            raise RuntimeError("Stdout piping failed")
        yield linter_process.stdout


def run_linter(
    cmdline: str, root: Path, paths: Set[Path], revrange: RevisionRange, use_color: bool
) -> int:
    """Run the given linter and print linting errors falling on changed lines

    :param cmdline: The command line for running the linter
    :param root: The common root of all files to lint
    :param paths: Paths of files to check, relative to ``root``
    :param revrange: The Git revision rango to compare
    :return: The number of modified lines with linting errors from this linter

    """
    _require_rev2_worktree(revrange.rev2)
    if not paths:
        return 0
    error_count = 0
    edited_linenums_differ = EditedLinenumsDiffer(root, revrange)
    missing_files = set()
    with _check_linter_output(cmdline, root, paths) as linter_stdout:
        prev_path, prev_linenum = None, 0
        for line in linter_stdout:
            (
                path_in_repo,
                linter_error_linenum,
                location,
                description,
            ) = _parse_linter_line(line, root)
            if (
                path_in_repo is None
                or path_in_repo in missing_files
                or linter_error_linenum == 0
            ):
                continue
            if path_in_repo.suffix != ".py":
                logger.warning(
                    "Linter message for a non-Python file: %s %s",
                    location,
                    description,
                )
                continue
            try:
                edited_linenums = edited_linenums_differ.compare_revisions(
                    path_in_repo, context_lines=0
                )
            except FileNotFoundError:
                logger.warning("Missing file %s from %s", path_in_repo, cmdline)
                missing_files.add(path_in_repo)
                continue
            if linter_error_linenum in edited_linenums:
                if path_in_repo != prev_path or linter_error_linenum > prev_linenum + 1:
                    print()
                prev_path, prev_linenum = path_in_repo, linter_error_linenum
                print(colorize(location, "lint_location", use_color), end=" ")
                print(colorize(description, "lint_description", use_color))
                error_count += 1
    return error_count


def run_linters(
    linter_cmdlines: List[str],
    root: Path,
    paths: Set[Path],
    revrange: RevisionRange,
    use_color: bool,
) -> int:
    """Run the given linters on a set of files in the repository

    :param linter_cmdlines: The command lines for linter tools to run on the files
    :param root: The root of the relative paths
    :param paths: The files to check, relative to ``root``. This should only include
                  files which have been modified in the repository between the given Git
                  revisions.
    :param revrange: The Git revisions to compare
    :return: Total number of linting errors found on modified lines

    """
    # 10. run linter subprocesses for all edited files (10.-13. optional)
    # 11. diff the given revision and worktree (after isort and Black reformatting)
    #     for each file reported by a linter
    # 12. extract line numbers in each file reported by a linter for changed lines
    # 13. print only linter error lines which fall on changed lines
    return sum(
        run_linter(linter_cmdline, root, paths, revrange, use_color)
        for linter_cmdline in linter_cmdlines
    )
